"This module implements generation of paper keys (used by validation)."
from __future__ import print_function

__author__ = "Anders Logg (logg@simula.no)"
__date__ = "2009-06-30 -- 2009-06-30"
__copyright__ = "Copyright (C) 2009 Anders Logg"
__license__  = "GNU GPL version 3 or any later version"

from publish.common import pstr

def generate_keys(papers):
    "Generate keys for all papers (if not already existing)"

    # Set of used keys
    used_keys = set()

    # Iterate over all papers
    for paper in papers:

        # Skip paper if key is already present
        if "key" in paper:
            continue

        # Generate key
        key = _generate_key(paper, used_keys)
        key = _filter_key(key)
        used_keys.add(key)

        # Set key
        paper["key"] = key
        print("Adding key for paper %s: %s" % (pstr(paper), key))

    return papers

def _generate_key(paper, used_keys):
    """Generate key for paper. The following naming scheme is used:

      FooBar2007
      FooBar2009a
      FooBar2009b

    The letter prefix is always used even if there is only one paper
    for the authors that year (so keys don't need to be updated if
    a new paper is added).

    If there are more than two authors, EtAl is added.

    Keys for some categories are formatted differently:

      courses: <code>.<year>
      talks:   FooBar<date> (with any "-" removed to work with LaTeX)
    """

    # Special case: courses
    if paper["category"] == "courses":
        return paper["code"] + "." + paper["year"]

    # Generate basic key
    last_names = [author.split(" ")[-1] for author in paper["author"]]
    key = "".join(last_names[:min(len(last_names), 2)])
    if len(last_names) > 2:
        key += "EtAl"
    if paper["category"] == "talks" and "date" in paper :
        key += paper["date"].replace("-", "")
    else:
        key += paper["year"]

    # Add suffix until we find something unique
    letters = "abcdefghijklmnopqrstuvxyz"
    for letter in letters:
        if not (key + letter) in used_keys:
            key = key + letter
            break

    # Check that we found a unique key
    if key in used_keys:
        raise RuntimeError("""Too many papers, out of suffix letter. This is a known bug/feature which should
be easy to fix.""")

    return key

def _filter_key(key):
    "Filter key, removing special characters."

    key = key.replace("{", "")
    key = key.replace("}", "")
    key = key.replace("\\", "")

    return key



"This module implements output for HTML."
from builtins import range

__author__ = "Anders Logg <logg@simula.no>"
__date__ = "2009-07-31 -- 2009-08-18"
__copyright__ = "Copyright (C) 2009 Anders Logg"
__license__  = "GNU GPL version 3 or any later version"

from publish import config

def write(papers, sort_func=None):
    "Format the given list of papers in HTML format."

    text = ""
    prefix = ""

    # Get the categories.
    # Assume unique names
    categories = config.get("categories")
    category_headings = config.get("category_headings")

    # Get formatting rule
    html_format = config.get("html_format")

    # Get PDF directory
    pdf_dir = config.get("pdf_dir")

    # Iterate over categories
    current_paper = 0
    for category in categories:

        # Extract papers in category
        category_papers = [paper for paper in papers if paper["category"] == category]
        if len(category_papers) == 0:
            continue

        # Sort the list
        if sort_func is not None :
          category_papers.sort(key=sort_func)

        # Add internal links to each category
        if config.get("html_add_internal_links") :
            prefix += "<li><a href=\"#%s_id_%s\">%s</a></li>\n" % (config.get("html_class_prefix"),
                                                                 category,
                                                                 category_headings[category])

        # Write category
        text += '<h2 id="%s_id_%s">%s</h2>\n\n' % ( config.get("html_class_prefix"),
                                                    category,
                                                    category_headings[category])
        text += "<ol class=\"%s_list\">\n\n" %  config.get("html_class_prefix")

        # Iterate over papers in category
        for paper in category_papers:

            # Format paper entry
            paper_entry = html_format[category](paper)

            # Filter entry from special characters
            paper_entry = _filter(paper_entry)

            # Set directory to papers (a bit of a hack)
            paper_entry = paper_entry.replace("papers/", pdf_dir + "/")

            # Write entry for paper
            text += "<li class=\"publish_item\">\n" + paper_entry + "</li>\n"

        # Write end of list
        text += "</ol>\n"

    prefix = "<ul>" + prefix + "</ul>\n"

    return prefix+"\n"+text

def _filter(s):
    "Filter string for special characters."

    # List of replacements
    replacements = [("--", "&ndash;"),
                    ("\\ae", "ae"),
                    ("$", ""),
                    ("\\mathrm", ""),
                    ('\\aa', '&aring;'),
                    ('\\AA', '&Aring;'),
                    ('\\"a', '&auml;'),
                    ('\\"A', '&Auml;'),
                    ('\\"o', '&ouml;'),
                    ('\\"O', '&Ouml;'),
                    ('\\o',  '&oslash;'),
                    ('\\O',  '&Oslash;'),
                    ('\\&',   '&amp;'),
                    ('\\textonesuperior', "<sup>1</sup>"),
                    ('\\textendash', "&ndash;"),
                    ('\\textemdash', "&mdash;"),
                    ('\\textasciiacute', "&acute;"),
                    ('\\\'a', "&acute;"),
                    ('\\"u', "&uuml;"),
                    ('\\epsilon', "&epsilon;"),
                    ('\\\'\i', "&iacute;"),
                    ('\\textquoteright', "&apos;"), # Is this correct?
                    ('\\textquoteleft', "&apos;"), # Is this correct?
                    ("\\'", "&apos;"),
                    ("\l", "&#0322;"),
                    ("\\.Z", "&#379;"),
                    ("\\beta", "&beta;"),
                    ("\\r a", "&aring;"), # Note that order matters here
                    ("\\r", "&aring;"),   # and here
                    ("\`a", "&agrave;")
                    ]

    # Sort wrt to length of string
    replacements.sort(key=len, reverse=True)

    # Remove { }
    while (True) :
        found = False
        for i in range(len(s)) :
            if s[i] in ["{", "}"] and (i == 0 or s[i-1] != "\\") :
                found = True
                s = s[:i] + s[i+1:]
                break
        if not found :
            break

    # Iterate over replacements
    for (a, b) in replacements:
        s = s.replace(a, b)

    return s



"This module implements output for pdf."
from __future__ import absolute_import

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-25 -- 2008-11-25"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Modified by Anders Logg 2011
# Last changed: 2012-06-02

from tempfile import mkstemp
from os import system

from publish import config
from . import latex

def write(papers, sort_func=None):
    "Format the given list of papers in the pdf format."

    # Start of pdf paper
    latex_text = ""
    latex_text += "\\documentclass[11pt]{article}\n"
    latex_text += "\\usepackage{textcomp}\n"
    latex_text += "\\usepackage{a4wide}\n"

    # Because of some font issues in latex, we need to include
    # this package if we are going to nest \textbf and and \textsc
    # \textbf is only used if we are going to mark specific authors
    if config.get("use_textsc") and len(config.get("mark_author")) > 0 :
        latex_text += "\\usepackage[T1]{fontenc}\n"
    latex_text += "\\usepackage{url}\n"
    latex_text += "\\begin{document}\n"

    # Get headline
    headline = config.get("headline")
    if headline == "": headline = "Publications"

    # Write headline
    repeat_headline = config.get("repeat_headline")
    if repeat_headline:
        latex_text += "\\renewcommand \\refname{%s}\n\n" % headline
    else:
        latex_text += "\\renewcommand \\refname{\\vspace{-0.7cm}}\n\n"
        latex_text += "\\section*{%s}\n\n" % headline

    # LaTeX output
    latex_text += latex.write(papers, sort_func)

    # End of pdf paper
    latex_text += "\\end{document}"

    # Write string to LaTeX file
    try:
        (handle, latex_filename) = mkstemp(".tex")
        latex_file = open(latex_filename, "w")
        latex_file.write(latex_text)
        latex_file.close()
    except:
        raise RuntimeError("Unable to generate intermediate LaTeX code for PDF output.")

    # FIXME: Specifying /tmp is platform-specific
    # FIXME: Remove temporary files

    # Run pdflatex on LaTeX file
    system("pdflatex -output-directory /tmp %s" % latex_filename)

    # Read PDF file
    try:
        pdf_filename = ".".join(latex_filename.split(".")[:-1]) + ".pdf"
        pdf_file = open(pdf_filename, "r")
        pdf_text = pdf_file.read()
        pdf_file.close()
    except:
        raise RuntimeError("Unable to read generated PDF output.")

    # Show PDF
    if config.get("view_pdf"):
        system("%s %s" % (config.get("pdf_viewer"), pdf_filename))

    return pdf_text




# Should all parsers (ie. Bibtex and Publish) share syntax error exception class

# This class is a bit simplistic at the moment. It should probably take 
# filename, line number and error message as separate arguments.
class ParseException(Exception) :
  pass



"""This module provides functionality for reading and writing papers
from the database."""
from __future__ import print_function
from __future__ import absolute_import

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-12-12 -- 2008-12-12"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

import shutil
import os.path
import time

from .formats import pub
from publish import config
from .common import is_valid

def read_database(database_filename=None):
    "Read papers from database"

    # Get location of database file
    if database_filename is None:
        database_filename = config.get("database_filename")

    # Open and read database
    try:
        file = open(database_filename, "r")
        text = file.read()
        file.close()
    except Exception as e:
        return []

    # Parse file
    database_papers = pub.read(text)

    return database_papers

def save_database(merged_papers):
    "Save to database and make a backup copy if needed"

    database_filename = config.get("database_filename")

    # Generate text to be written to file
    text = pub.write(merged_papers)

    print("")

    # Make backup copy if needed (file size of generated file is different from the current)
    # TODO: Register if changes has been made and write backup file based on that
    #       (instead of just comparing file sizes)
    if os.path.isfile(database_filename) and len(text) != os.path.getsize(database_filename):
        backup_filename = database_filename + ".bak"
        print('Saving backup copy of database to file "%s"' % backup_filename)
        try:
            shutil.copyfile(database_filename, backup_filename)
        except:
            raise RuntimeError("Unable to create backup copy of database")

    # Open and read file
    print('Saving database to file "%s"' % database_filename)
    try:
        file = open(database_filename, "w")
        file.write(text)
    except UnicodeEncodeError as e:
        try:
            file.write(text.encode('utf-8'))
        except Exception as e:
            raise RuntimeError('Unable to save database to file "%s"\n%s' % (database_filename, str(e)))

    file.close()

def save_invalid_papers(papers):
    "Save invalid papers to file"

    # Extract invalid papers
    invalid_papers = []
    for paper in papers:
        if not is_valid(paper):
            invalid_papers.append(paper)

    # Don't save if there are no invalid papers
    if len(invalid_papers) == 0:
        return

    # Generate filename
    date = time.strftime("%Y%m%d-%H:%m:%S")
    invalid_filename = config.get("invalid_filename_prefix") + "-" + date + ".pub"

    # Write to file
    text = pub.write(invalid_papers)
    print('Saving invalid papers to "%s".' % invalid_filename)
    try:
        file = open(invalid_filename, "w")
        file.write(text)
        file.close()
    except:
        raise RuntimeError('Unable to save invalid papers to file "%s"' % invalid_filename)



"This module implements data import from different file formats."
from __future__ import print_function
from __future__ import absolute_import

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-13 -- 2008-12-12"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

from .formats import bibtex, latex, pub
from .validation import validate_papers
from .merging import merge_papers
from publish import config
from .log import print_summary
from .database import read_database, save_database, save_invalid_papers

def import_file(filename, filters=[]):
    "Import data from file into database."

    # Read and validate database
    database_papers = read_database()
    if not config.get("simple_import") :
        (database_papers, invalid_database_papers) = validate_papers(database_papers)
    else :
        invalid_database_papers = []

    # Read imported file
    imported_papers = _read_file(filename)

    # Assign attributes to imported papers
    _assign_attributes(imported_papers, filters)

    # Validate imported file
    if not config.get("simple_import") :
        (imported_papers, invalid_imported_papers) = validate_papers(imported_papers)
    else :
        invalid_imported_papers = []
    print("")
    
    if len(invalid_imported_papers) > 0:
        print('Imported %d paper(s) from "%s" (and found %d invalid paper(s)).' % \
              (len(imported_papers), filename, len(invalid_imported_papers)))
    else:
        print('Imported %d paper(s) from "%s".' % (len(imported_papers), filename))

    # Merge papers
    if config.get("simple_import") :
        merged_papers = database_papers + imported_papers
    else :
        merged_papers = merge_papers(database_papers, imported_papers)

    # Print summary
    print_summary(merged_papers)

    # Save papers to database
    save_database(merged_papers)
    save_invalid_papers(invalid_database_papers + invalid_imported_papers + merged_papers)

def _read_file(filename):
    "Read papers from file"

    # Get the filename suffix
    suffix = filename.split(".")[-1]

    # Choose format based on suffix
    if suffix in ("bib", "bibtex"):
        read = bibtex.read
    elif suffix == "pub":
        read = pub.read
    else:
        raise RuntimeError("Unknown file format.")

    # Open and read file
    try:
        file = open(filename, "r")
        text = file.read()
        file.close()
    except:
        raise RuntimeError('Unable to open file "%s"' % filename)

    # Parse file
    imported_papers = read(text)

    return imported_papers

def _assign_attributes(imported_papers, filters):
    "Assign attributes to imported file"

    for paper in imported_papers:
        for attribute in filters:
            value, should_match = filters[attribute]
            if should_match:
                paper[attribute] = value
            else:
                RuntimeError, "Illegal assignment of attribute: %s!=%s" % (attribute, value)



"This module implements data import from different file formats."
from __future__ import print_function
from builtins import str
from builtins import range

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-27 -- 2008-11-08"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Last changed: 2012-08-31

# Modified by Benjamin Kehlet 2012

from publish import config
from publish.common import pstr, is_duplicate, is_allowed_duplicate
from publish.algorithms import distance
from publish.interaction import ask_user_alternatives
from publish.formats import pub

def merge_papers(papers_0, papers_1):
    "Merge two lists of papers into one common list"

    print("")
    print("Merging papers")
    print("--------------")
    print("")
    print("Need to merge %d + %d = %d papers." % \
          (len(papers_0), len(papers_1), len(papers_0) + len(papers_1)))
    print("")

    # Concatenate lists and then check every paper against all others
    papers = papers_0 + papers_1

    merged_papers = process_duplicates(papers)

    return merged_papers


def process_duplicates(papers) :
    " Search for duplicates, ask user what to do, return processed list"

    merged_papers = []

    for paper in papers:

        # Look for close match
        (matching_paper, matching_position) = _find_matching_paper(paper, merged_papers)

        # Add paper if no match
        if matching_paper is None:
            merged_papers.append(paper)
            continue

        # Ignore exact duplicates
        if _exact_match(paper, matching_paper):
            print("Found exact duplicate paper %s, skipping." % pstr(paper))
            continue

        # Add paper if matching paper marked as duplicate
        # NOTE: The attribute "duplicate" is now superseded by "allowed_duplicates". 
        # "duplicate" will be read and respected when merging, but not written

        if is_duplicate(matching_paper):
            print("Warning: Attribute 'duplicate' is deprecated. Use 'allowed_duplicates' and specify keys of duplicates.")
            print("Found close match with allowed duplicate for paper %s, keeping paper.\n" % pstr(paper))
            merged_papers.append(paper)
            continue

        if is_allowed_duplicate(paper, matching_paper) :
            print("Found close match with allowed duplicate for paper %s, keeping paper.\n" % pstr(paper))
            merged_papers.append(paper)
            continue


        # Remove matching paper (so we can add back the merged paper)
        del merged_papers[matching_position]

        # Merge paper with matching paper
        print("Found close match between %s and %s, merging papers." % (pstr(paper), pstr(matching_paper)))        
        merged_paper = _merge_papers(matching_paper, paper)
        for p in merged_paper:
            merged_papers.append(p)

    return merged_papers

def _find_matching_paper(paper, merged_papers):
    "Find paper that is so similar that it is probably intended to be the same"

    min_distance = 1.0
    matching_paper = None
    matching_position = None
    matching_distance = config.get("matching_distance_strong")

    # Get venue type
    category = paper["category"]
    category_venues = config.get("category_venues")
    # "journal", "booktitle", etc
    venue_type = category_venues[category] if "category" in category_venues else None 

    # Compare against all papers
    for i in range(len(merged_papers)):

        p = merged_papers[i]

        # Handle the special case where one of the papers is a talk
        if ((paper["category"] == "talks" or p["category"] == "talks") and
            paper["category"] != p["category"] and config.get("talks_dont_duplicate")) :
            continue

        # Compute venue distance
        if venue_type is None or not p["category"] == category:
            d_venue = 0
        else:
            d_venue = distance(paper[venue_type].lower(), p[venue_type].lower())

        # Compute title distance
        d_title = distance(paper["title"].lower(), p["title"].lower())

        # Compute overall distance
        d = max(d_venue, d_title)

        # Check for match
        if d < matching_distance and d < min_distance:
            min_distance = d
            matching_paper = p
            matching_position = i

    return (matching_paper, matching_position)

def _exact_match(paper0, paper1):
    "Check if papers match exactly."

    ignores = ["duplicate", "invalid", "allowed_duplicates"]

    attributes = [attribute for attribute in paper0 if not attribute in ignores] + \
                 [attribute for attribute in paper1 if not attribute in ignores]

    for attribute in attributes:
        if not attribute in paper0 or not attribute in paper1:
            return False
        if not paper0[attribute] == paper1[attribute]:
            return False

    return True

def _merge_papers(paper0, paper1):
    "Merge papers"

    if "key" in paper0 and "title" in paper0:
        print("  " + paper0["key"] + ": " + paper0["title"])
    if "key" in paper1 and "title" in paper1:
        print("  " + paper1["key"] + ": " + paper1["title"])

    # Get all attributes to check
    common_attributes = []
    for attribute in paper0:
        if attribute in paper1:
            common_attributes.append(attribute)
    for attribute in paper1:
        if attribute in paper0 and attribute not in common_attributes:
            common_attributes.append(attribute)

    # Sort the attributes to make sure output is readable and consistent across
    # runs (otherwise regression tests will fail)
    common_attributes.sort()
    print(common_attributes)

    # Check all common attributes
    merged_paper = {}
    identical = True
    for attribute in common_attributes:
        value0 = paper0[attribute]
        value1 = paper1[attribute]
        if value0 == value1:
            merged_paper[attribute] = value0
        elif value0 == "":
            merged_paper[attribute] = value1
        elif value1 == "":
            merged_paper[attribute] = value0
        else:
            identical = False
            while True:
                alternative = ask_user_alternatives('  Attribute "%s" differs, what should I do?' % attribute,
                                                    ('Keep both papers (marking them as allowed duplicates).',
                                                     'Ignore papers (marking them as invalid).',
                                                     'Keep first paper (%s) and ignore second paper (%s)' % \
                                                     (pstr(paper0), pstr(paper1)),
                                                     'Keep second paper (%s) and ignore first paper (%s)' % \
                                                     (pstr(paper0), pstr(paper1)),
                                                     'Use attribute from first paper ("%s")' % str(value0),
                                                     'Use attribute from second paper ("%s")' % str(value1),
                                                     'Print diff.'))
                
                print("")
                if alternative == 0:
                    paper0["allowed_duplicates"] = [paper1["key"]] + (paper0["allowed_duplicates"] if "allowed_duplicates" in paper0 else [])
                    paper1["allowed_duplicates"] = [paper0["key"]] + (paper1["allowed_duplicates"] if "allowed_duplicates" in paper1 else [])

                    return [paper0, paper1]
                elif alternative == 1:
                    paper0["invalid"] = True
                    paper1["invalid"] = True
                    return [paper0, paper1]
                elif alternative == 2:
                    return [paper0]
                elif alternative == 3:
                    return [paper1]
                elif alternative == 4:
                    merged_paper[attribute] = value0
                    break
                elif alternative == 5:
                    merged_paper[attribute] = value1
                    break
                elif alternative == 6:
                    print(pub.write_diff(paper0, paper1))
                else:
                    raise RuntimeError("Unknown option.")

    if identical:
        print("  No conflicting attributes, merge ok.")
    
    # Add unique attributes from paper0
    for attribute in paper0:
        if not attribute in paper1:
            merged_paper[attribute] = paper0[attribute]
    for attribute in paper1:
        if not attribute in paper0:
            merged_paper[attribute] = paper1[attribute]

    return [merged_paper]



"Some common utility functions"
from __future__ import absolute_import
from builtins import str
from builtins import range

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-10 -- 2008-11-11"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Last modified: 2012-08-30
# Modified by Anders Logg 2012
# Modified by Benjamin Kehlet 2012

from .log import warning

from publish import config

def is_valid(paper):
    "Check if paper is valid"
    return not ("invalid" in paper and paper["invalid"])


def is_duplicate(paper):
    "Check if paper is a duplicate"
    return "duplicate" in paper and paper["duplicate"]


def is_allowed_duplicate(paper1, paper2) :
    "Check if papers are allowed duplicates"
    return (("allowed_duplicates" in paper1 and paper2["key"] in paper1["allowed_duplicates"]) or
           ("allowed_duplicates" in paper2 and paper1["key"] in paper2["allowed_duplicates"]))

def pstr(paper):
    "Return a simple string representation of the paper"

    s = "(%s) - %s" % (paper.get("key", "No key"),
                       paper.get("title", "No title"))

    if len(s) > 75 :
        s = s[:75]+"..."

    return s
    if "title" in paper:
        s = paper["title"]
        n = 75
        if len(s) > n:
            s = s[:n] + "..."
    else:
        s = "Unknown"
    return s

def split_with_quotes(text) :
    "Split text into a tuple of words respecting quotes and (non-escaped) curly brackets"

    splitted = []
    last_split = 0
    in_quotes = False
    brackets_level = 0

    text = text.strip()
    i = 0
    while i < len(text) :
        if text[i].isspace() and not in_quotes and brackets_level == 0 :
            splitted.append(text[last_split:i])
            i += 1
            while text[i].isspace() and i < len(text) :
                i += 1

            last_split = i
        elif text[i] == "}" and (i == 0 or text[i-1] != "\\") :
            brackets_level = max(0, brackets_level-1)
        elif text[i] == "{" and (i == 0 or text[i-1] != "\\") :
            brackets_level += 1
        elif text[i] == "\"" and (i == 0 or text[i-1] != "\\") :
            in_quotes = not in_quotes
        i += 1

    if i > last_split :
        splitted.append(text[last_split:i])

    if brackets_level > 0 :
        warning("Mismatched brackets")

    return tuple(splitted)


def short_author(author):
    "Abbreviate author name with initials"

    words = split_with_quotes(author)

    # This is an alternative, but leaves empty string
    # in the result list
    # words = author.split('\"')

    new_words = []
    for word in words[:-1]:
        if word == "":
            continue
        new_words.append(find_initials(word, author))
    return " ".join(new_words) + " " + words[-1]

def find_initials(name, author):
    "Generate initial string for name"

    if name == "":
        return ""

    # Call recursively for hyphen in name
    if "-" in name:
        return "-".join([find_initials(part, author) for part in name.split("-")])

    # Simple case: just a plain character
    if name[0].isalpha():
        return name[0] + "."

    # Nasty case: begins with "{"
    for i in range(len(name)):
        if name[i] == "}":
            return name[:i + 1] + "."

    # Unhandled case, just return full name
    warning("Unable to abbreviate author name, too complex: " + str(author))
    return name

def ordered_attributes(paper, ignores=[]):
    "Return list of ordered attributes for paper"
    attributes = []
    for attribute in config.get("ordered_attributes"):
        if attribute in paper and not attribute in ignores:
            attributes.append(attribute)
    # add remaining attributes in alphabetical order
    paper_attributes = [a for a in paper]
    paper_attributes.sort()
    for attribute in paper_attributes:
        if not attribute in attributes and not attribute in ignores:
            attributes.append(attribute)
    return attributes



"This module controls formatting for LaTeX and HTML."
from builtins import str

__author__    = "Anna Logg (anna@loggsystems.se)"
__date__      = "2008-11-17 -- 2009-07-31"
__copyright__ = "Copyright (C) 2008-2009 Anna Logg"
__license__   = "GNU GPL version 3 or any later version"

# Modified by Anders Logg, 2009.
# Modified by Benjamin Kehlet, 2010-2014.

from publish.common import short_author
from publish import config

#------------------------------------------------------------------------------
# LaTeX formatting
#------------------------------------------------------------------------------

def latex_format_articles(paper):
    "Return string for article in LaTeX format"

    text = []

    # authors
    text.append("%s. " %_latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    text.append("%s.\n" % paper["title"])

    # journal
    journal_in_italic = "\\textit{%s}" % paper["journal"]
    text.append("%s" % _format_venue(journal_in_italic, paper["journal"], paper))

    # volume
    if "volume" in paper and paper["volume"].strip() :
        text.append(", vol. %s" % paper["volume"])
        if "number" in paper and paper["number"].strip() :
            text.append("(%s)" % paper["number"])

    # pages
    if "pages" in paper and paper["pages"].strip() :
        text.append(", pp. %s" %  paper["pages"])

    # year
    text.append(", %s." % paper["year"])

    return "".join(text)

def latex_format_books(paper):
    "Return string for book in LaTeX format"
    values = []

    # authors
    values.append("%s.\n" % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    values.append(paper["title"])

    # edition
    if "edition" in paper :
        values += [", %s edition" % paper["edition"]]

    # publisher
    values.append(", \\textit{%s}" % paper["publisher"])

    # year
    values.append(", %s." % paper["year"])

    return "".join(values)

def latex_format_edited(paper):
    "Return string for edited book in LaTeX format"
    values = []
    values += ["%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"]))]
    values += ["%s, " % paper["title"]]
    values += ["\\textit{%s}, " % paper["publisher"]]
    values += [paper["year"]+"."]
    return "\n".join(values)

def latex_format_chapters(paper):
    "Return string for chapter in LaTeX format"

    values = []

    # authors
    values.append("%s. " %_latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    values.append(paper["title"])

    # book title
    values.append(". %s, " % _format_venue("\\textit{%s}" % paper["booktitle"], paper["booktitle"], paper, add_in=True))

    #editor
    values.append("edited by %s, " % _latex_get_authors_string(paper["editor"]))

    # publisher
    values.append("%s, " % paper["publisher"])

    #if "chapter" in paper: values += [" chapter %s" % paper["chapter"]]
    #if "pages" in paper: values += [" pp. %s" % _format_pages(paper["pages"])]

    values.append(" %s." % paper["year"])
    return "".join(values)

def latex_format_proceedings(paper):
    "Return string for proceeding in LaTeX format"
    values = []

    # authors
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    values.append(paper["title"])

    # book title
    values.append(". %s, " % _format_venue("\\textit{%s}" % paper["booktitle"], paper["booktitle"], paper, add_in=True))

    # year
    values.append(paper["year"])

    return "".join(values)

def latex_format_reports(paper):
    "Return string for report in LaTeX format"
    values = []

    # authors
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    values.append(paper["title"])

    # institution
    values.append(", %s, " % paper["institution"])

    # year
    values.append(paper["year"])
    return "".join(values)

def latex_format_manuals(paper):
    "Return string for manual in LaTeX format"
    values = []

    # authors
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    values.append(paper["title"])


    if "year" in paper: values += [", %s" % paper["year"]]
    return "".join(values)

def latex_format_theses(paper):
    "Return string for thesis in LaTeX format"
    values = []

    # authors
    values.append("%s. " %_latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    values.append(paper["title"])

    #thesis type
    values.append(", %s" % config.get("thesistype_strings")[paper["thesistype"]])

    # school and year
    values.append(", %s, %s." % (paper["school"], paper["year"]))

    return "".join(values)

def latex_format_courses(paper):
    "Return string for course in LaTeX format"
    values = []
    values += [_latex_format_authors(_latex_get_authors_string(paper["author"]))]
    values += ["\\textit{%s}" % paper["title"]]
    values += ["(" + paper["code"] + ")"]
    values += [paper["institution"]]
    values += [paper["year"]]
    return _latex_join(values)

def latex_format_talks(paper):
    "Return string for talk in LaTeX format"
    values = []

    # author
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    #title
    values.append(paper["title"])

    # meeting
    if "meeting" in paper :
        values.append(", %s" % paper["meeting"])

    #year
    values.append(", %s." % paper["year"])

    return "".join(values)

def latex_format_posters(paper):
    "Return string for poster in LaTeX format"
    values = []

    # author
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    #title
    values.append(paper["title"])

    # meeting
    if "meeting" in paper :
        values.append(", %s" % paper["meeting"])

    #year
    values.append(", %s." % paper["year"])

    return "".join(values)

def latex_format_publicoutreach(paper):
    "Return string for public outreach in LaTeX format"
    values = []

    # author
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    #title
    values.append(paper["title"])

    # meeting
    if "meeting" in paper :
        values.append(", %s" % paper["meeting"])

    #year
    values.append(", %s." % paper["year"])

    return "".join(values)

def latex_format_preprint(paper):
    "Return string for preprint in LaTeX format"
    values = []

    # authors
    values.append("%s. " % _latex_format_authors(_latex_get_authors_string(paper["author"])))

    # title
    values.append(paper["title"])

    # title
    values.append(", %s" % paper["year"])
    return "".join(values)

def latex_format_misc(paper):
    "Return string for misc in LaTeX format"
    values = []
    values += [_latex_format_authors(_latex_get_authors_string(paper["author"]))]
    values += ["\\textit{%s}" % paper["title"]]
    if "booktitle" in paper: values += ["in \\textit{%s}" % paper["booktitle"]]
    if "howpublished" in paper: values += [paper["howpublished"]]
    if "meeting" in paper: values += [paper["meeting"]]
    if "thesistype" in paper: values += [config.get("thesistype_strings")[paper["thesistype"]]]
    if "school" in paper: values += [paper["school"]]
    if "chapter" in paper: values += ["chapter %s" % paper["chapter"]]
    if "volume" in paper: values += ["vol. %s" % paper["volume"]]
    if "pages" in paper: values += ["pp. %s" % _latex_format_pages(paper["pages"])]
    if "year" in paper: values += [paper["year"]]
    return _latex_join(values)

def _latex_join(values):
    "Join values for LaTeX entry"
    return ",\n".join(values) + ".\n"

latex_format = {"articles"      : latex_format_articles,
                "books"         : latex_format_books,
                "edited"        : latex_format_edited,
                "chapters"      : latex_format_chapters,
                "proceedings"   : latex_format_proceedings,
                "refproceedings": latex_format_proceedings,
                "reports"       : latex_format_reports,
                "manuals"       : latex_format_manuals,
                "theses"        : latex_format_theses,
                "courses"       : latex_format_courses,
                "talks"         : latex_format_talks,
                "posters"       : latex_format_posters,
                "publicoutreach": latex_format_publicoutreach,
                "preprint"      : latex_format_preprint,
                "misc"          : latex_format_misc}

#------------------------------------------------------------------------------
# HTML formatting
#------------------------------------------------------------------------------

def html_format_articles(paper):
    "Return string for article in HTML format"
    values = []

    # Title
    values.append(_html_format_title(paper))

    # Author
    values.append(_html_get_authors_string(paper["author"]))

    # Journal
    values.append('<span class="%s_item_journal">%s</span>' % (config.get("html_class_prefix"), _format_venue(paper["journal"], paper["journal"], paper)))

    # Volume
    if "volume" in paper: values.append('<span class="%s_item_volum">vol. %s</span>' %  (config.get("html_class_prefix"), paper["volume"]))

    # Pages
    if "pages" in paper: values.append('<span class="%s_item_pages">pp. %s</span>' % (config.get("html_class_prefix"),
                                                                                      _html_format_pages(paper["pages"])))

    # Year
    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))

    # DOI
    #if "doi" in paper: values.append('[<a href="http://dx.doi.org/%s">DOI:%s</a>]' % (paper["doi"], paper["doi"]))
    if "doi" in paper: values.append('[<a href="http://dx.doi.org/%s">DOI</a>]' % paper["doi"])

    # arXiv
    #if "arxiv" in paper: values.append('[<a href="http://arxiv.org/abs/%s">arXiv:%s</a>]' % (paper["arxiv"], paper["arxiv"]))
    if "arxiv" in paper: values.append('[<a href="http://arxiv.org/abs/%s">arXiv</a>]' % paper["arxiv"])

    return _html_join(values)

def html_format_books(paper):
    "Return string for book in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += ['<span class="%s_item_publisher">%s</span>' % (config.get("html_class_prefix"), paper["publisher"])]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_edited(paper):
    "Return string for edited book in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += ['<span class="%s_item_publisher">%s</span>' % (config.get("html_class_prefix"), paper["publisher"])]
    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))

    return _html_join(values)

def html_format_chapters(paper):
    "Return string for chapter in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += ['in <span class="%s_item_publisher">%s</span>' % (config.get("html_class_prefix"), paper["booktitle"])]
    values += [_html_format_editors(paper["editor"])]
    values += ['<span class="%s_item_publisher">%s</span>' % (config.get("html_class_prefix"), paper["publisher"])]
    if "chapter" in paper: values += ["chapter %s" % paper["chapter"]]
    if "pages" in paper: values += ["pp. %s" % _html_format_pages(paper["pages"])]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))

    return _html_join(values)

def html_format_proceedings(paper):
    "Return string for proceeding in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += ['in <span class="%s_item_booktitle">%s</span>' % (config.get("html_class_prefix"), paper["booktitle"])]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_reports(paper):
    "Return string for report in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += [paper["institution"]]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_manuals(paper):
    "Return string for manual in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    if "year" in paper: values += [paper["year"]]
    return _html_join(values)

def html_format_theses(paper):
    "Return string for thesis in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += [config.get("thesistype_strings")[paper["thesistype"]]]
    values += [paper["school"]]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_courses(paper):
    "Return string for course in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += [_html_get_authors_string(paper["author"])]
    values += [paper["institution"]]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_talks(paper):
    "Return string for talk in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    if "meeting" in paper :
        values += ['<span class="%s_item_meeting">%s</span>' % (config.get("html_class_prefix"), paper["meeting"])]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_posters(paper):
    "Return string for poster in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += ['<span class="%s_item_meeting">%s</span>' % (config.get("html_class_prefix"), paper["meeting"])]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)



def html_format_preprint(paper):
    "Return string for preprint in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    if "year" in paper: values += [paper["year"]]
    return _html_join(values)

def html_format_publicoutreach(paper):
    "Return string for public outreach in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    values += ['<span class="%s_item_meeting">%s</span>' % (config.get("html_class_prefix"), paper["meeting"])]

    values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def html_format_misc(paper):
    "Return string for misc in HTML format"
    values = []
    values += [_html_format_title(paper)]
    values += [_html_get_authors_string(paper["author"])]
    if "howpublished" in paper:
        howpublished = paper["howpublished"]
        if "http://" in howpublished and "<a href" not in values[0]:
            link = ("http://" + howpublished.split("http://")[-1]).strip()
            values[0] = '<a href="%s">%s</a>' % (link, values[0])
        else:
            values += [howpublished]
    if "booktitle" in paper: values += ["in <i>%s</i>" % paper["booktitle"]]
    if "meeting" in paper: values += [paper["meeting"]]
    if "thesistype" in paper: values += [config.get("thesistype_strings")[paper["thesistype"]]]
    if "school" in paper: values += [paper["school"]]
    if "chapter" in paper: values += ["chapter %s" % paper["chapter"]]
    if "volume" in paper: values += ["vol. %s" % paper["volume"]]
    if "pages" in paper: values += ["pp. %s" % _html_format_pages(paper["pages"])]
    if "year" in paper: values.append('<span class="%s_item_year">%s</span>' % (config.get("html_class_prefix"), paper["year"]))
    return _html_join(values)

def _html_format_title(paper):
    "Format title for HTML, with or without link to PDF file"

    if paper["category"] == "courses":
        title = "%s (%s)" % (paper["title"], paper["code"])
    else:
        title = paper["title"]

    if "pdf" in paper and not paper["pdf"] == "missing":
        return '<span class="publish_item_title"><a href=\"%s\">%s</a></span>' % (paper["pdf"], title)
    else:
        return '<span class="publish_item_title">%s</span>' % title

def _html_format_editors(authors):
    "Convert editor tuple to author string"
    return '<span class="publish_item_editors">Edited by %s</span>' % _html_get_authors_string(authors)

def _html_get_authors_string(authors):
    "Convert author tuple to author string"
    authors = [_html_mark_author(author, short_author(author).strip()) for author in authors]
    if len(authors) == 1:
        str = authors[0]
    else :
        if authors[-1] == "others":
            str =  ", ".join(authors[:-1]) + " et al."
        else:
            str = ", ".join(authors[:-1]) + " and " + authors[-1]

    return '<span class="%s_item_authors">%s</span>' % (config.get("html_class_prefix"), str)

def _html_mark_author(author, text) :
  "Mark the text with bold face if author is in the list of marked authors"

  if author.strip() in config.get("mark_author") :
    return "<strong>%s</strong>" % text

  else :
    return text

def _html_format_pages(pages):
    "Format pages"
    if "--" in pages: return pages.replace("--", "&mdash;")
    else :            return pages.replace("-", "&mdash;")

def _html_join(values):
    "Join values for HTML entry"
    entry = "<br>\n".join(values[:2]) + "<br>\n" + ", ".join(values[2:]) + "\n"
    entry = entry.replace("{", "")
    entry = entry.replace("}", "")
    return entry

html_format = {"articles"      : html_format_articles,
               "books"         : html_format_books,
               "edited"        : html_format_edited,
               "chapters"      : html_format_chapters,
               "proceedings"   : html_format_proceedings,
               "refproceedings": html_format_proceedings,
               "reports"       : html_format_reports,
               "manuals"       : html_format_manuals,
               "theses"        : html_format_theses,
               "courses"       : html_format_courses,
               "talks"         : html_format_talks,
               "posters"       : html_format_posters,
               "preprint"      : html_format_preprint,
               "publicoutreach": html_format_publicoutreach,
               "misc"          : html_format_misc}

#------------------------------------------------------------------------------
# reSt formatting
#------------------------------------------------------------------------------

def rst_format_articles(paper):
    "Return string for article in reSt format"
    values = []

    # Author
    values.append(_rst_get_authors_string(paper))

    # Title
    values.append(_rst_format_title(paper))

    # Journal
    values.append(_format_venue(paper["journal"], paper["journal"], paper))

    # Volume
    if "volume" in paper:
        vol = paper["volume"]
        if "number" in paper :
            vol += "(%s)" % paper["number"]
        values.append(vol)

    # Pages
    if "pages" in paper:
        values.append(_rst_format_pages(paper["pages"]))

    # DOI
    if "doi" in paper:
        values.append(_rst_format_doi(paper["doi"]))

    # arXiv
    if "arxiv" in paper:
        values.append(_rst_format_arxiv(paper["arxiv"]))

    return _rst_join(values)

def rst_format_books(paper):
    "Return string for book in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["publisher"]]
    if "doi" in paper: values.append(_rst_format_doi(paper["doi"]))
    return _rst_join(values)

def rst_format_edited(paper):
    "Return string for edited book in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["publisher"]]
    return _rst_join(values)

def rst_format_chapters(paper):
    "Return string for chapter in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["booktitle"]]
    #values += [_rst_format_editors(paper)]
    values += [paper["publisher"]]
    if "chapter" in paper: values += ["Chapter %s" % paper["chapter"]]
    if "pages" in paper: values += ["pp. %s" % _rst_format_pages(paper["pages"])]
    return _rst_join(values)

def rst_format_proceedings(paper):
    "Return string for proceeding in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["booktitle"]]
    return _rst_join(values)

def rst_format_reports(paper):
    "Return string for report in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["institution"]]
    values += [paper["number"]]
    return _rst_join(values)

def rst_format_manuals(paper):
    "Return string for manual in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    if "year" in paper: values += [paper["year"]]
    return _rst_join(values)

def rst_format_theses(paper):
    "Return string for thesis in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [config.get("thesistype_strings")[paper["thesistype"]]]
    values += [paper["school"]]
    values += [paper["year"]]
    return _rst_join(values)

def rst_format_courses(paper):
    "Return string for course in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["institution"]]
    values += [paper["year"]]
    return _rst_join(values)

def rst_format_talks(paper):
    "Return string for talk in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["meeting"]]
    values += [paper["year"]]
    return _rst_join(values)

def rst_format_posters(paper):
    "Return string for poster in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    values += [paper["meeting"]]
    values += [paper["year"]]
    return _rst_join(values)


def rst_format_misc(paper):
    "Return string for misc in reSt format"
    values = []
    values += [_rst_get_authors_string(paper)]
    values += [_rst_format_title(paper)]
    if "howpublished" in paper:
        howpublished = paper["howpublished"]
        values += [howpublished]
    if "booktitle" in paper: values += ["in *%s*" % paper["booktitle"]]
    if "meeting" in paper: values += [paper["meeting"]]
    if "thesistype" in paper: values += [config.get("thesistype_strings")[paper["thesistype"]]]
    if "school" in paper: values += [paper["school"]]
    if "chapter" in paper: values += ["Chapter %s" % paper["chapter"]]
    if "volume" in paper: values += ["vol. %s" % paper["volume"]]
    if "pages" in paper: values += ["pp. %s" % _rst_format_pages(paper["pages"])]
    if "year" in paper: values.append(paper["year"])
    return _rst_join(values)

def _rst_format_title(paper):
    "Format title for reSt, with or without link to PDF file"
    if paper["category"] == "courses":
        title = "%s (%s)" % (paper["title"], paper["code"])
    else:
        title = paper["title"]
    return "*%s*" % title

def _rst_format_editors(paper):
    "Convert editor tuple to author string"
    return "Edited by %s" % _rst_get_authors_string(paper)

def _rst_get_authors_string(paper):
    "Convert author tuple to author string"
    authors = paper["author"]
    authors = [_rst_mark_author(author, short_author(author).strip()) \
                   for author in authors]
    if len(authors) == 1:
        str = authors[0]
    else :
        if authors[-1] == "others":
            str =  ", ".join(authors[:-1]) + " et al."
        else:
            str = ", ".join(authors[:-1]) + " and " + authors[-1]
    str = "**%s (%s)**" % (str, paper["year"])
    return str

def _rst_mark_author(author, text) :
  "Mark the text with bold face if author is in the list of marked authors"
  if author.strip() in config.get("mark_author") :
    return "**%s**" % text
  else:
    return text

def _rst_format_pages(pages):
    "Format pages"
    if "--" in pages:
        pages = pages.replace("--", "-")
    return "pp. %s" % pages

def _rst_format_doi(doi):
    "Format DOI"
    return "[`doi:%s <http://dx.doi.org/%s>`_]" % (doi, doi)

def _rst_format_arxiv(arxiv):
    "Format arXiv"
    return "[`arXiv:%s <http://arxiv.org/abs/%s>`_]" % (arxiv, arxiv)

def _rst_join(values):
    "Join values for reSt entry"
    entry = "* " + values[0] + ".\n  " + ",\n  ".join(values[1:]) + "." + "\n"
    entry = entry.replace("{", "")
    entry = entry.replace("}", "")
    return entry

rst_format = {"articles"      : rst_format_articles,
              "books"         : rst_format_books,
              "edited"        : rst_format_edited,
              "chapters"      : rst_format_chapters,
              "proceedings"   : rst_format_proceedings,
              "refproceedings": rst_format_proceedings,
              "reports"       : rst_format_reports,
              "manuals"       : rst_format_manuals,
              "theses"        : rst_format_theses,
              "courses"       : rst_format_courses,
              "talks"         : rst_format_talks,
              "posters"       : rst_format_posters,
              "misc"          : rst_format_misc}

#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

def _latex_mark_author(author, text) :
  "Mark the text with bold face if author is in the list of marked authors"

  if author.strip() in config.get("mark_author") :
    return "\\textbf{%s}" % text

  else :
    return text

def _latex_format_authors(author_string) :
  if config.get("use_textsc") :
    return "\\textsc{%s}" % author_string
  else :
    return author_string

def _latex_get_authors_string(authors):
    "Convert author tuple to author string"
    authors = [_latex_mark_author(author, short_author(author).strip()) for author in authors]
    if len(authors) == 1:
        return authors[0]
    if authors[-1] == "others":
        return ", ".join(authors[:-1]) + " et al."
    else:
        return ", ".join(authors[:-1]) + " and " + authors[-1]

def _latex_format_editors(authors):
    "Convert editor tuple to author string"
    return "Edited by " + _latex_get_authors_string(authors)

def _latex_format_pages(pages):
    "Format pages"
    if "--" in pages: return pages
    return pages.replace("-", "--")

def _format_venue(formatted_venue, venue, paper, add_in=False):
    "Format venue"
    status = paper.get("status", "published")
    if status == "published":
        if add_in:
            return "In " + formatted_venue
        else:
            return formatted_venue
    elif status == "accepted":
        return "Accepted for publication in " + formatted_venue
    elif status == "submitted":
        #if venue == "none":
        #    return "Submitted to journal for publication"
        #else:
        #    return "Submitted to " + formatted_venue
        return "Submitted to journal for publication"
    else:
        return "(%s)" % str(status)



"This module implements data export to different file formats."
from __future__ import print_function
from __future__ import absolute_import

__author__ = "Anders Logg (logg@simula.no)"
__date__ = "2010-01-04 -- 2010-01-04"
__copyright__ = "Copyright (C) 2010 Anders Logg"
__license__  = "GNU GPL version 3 or any later version"

from publish.importing import read_database
from .validation import validate_papers
from .filtering import filter_papers
from .log import print_summary
from publish import config

def print_file(filters=[]):
    "Print data to stdout"

    # Read database
    database_filename = config.get("database_filename")
    database_papers = read_database(database_filename)

    # Validate papers
    #(valid_papers, invalid_papers) = validate_papers(database_papers)
    valid_papers = database_papers

    # Filter papers
    filtered_papers = filter_papers(valid_papers, filters)

    # Print papers
    print()
    for paper in filtered_papers:
        _print_paper(paper)
        print()

    # Print summary
    print_summary(filtered_papers)

def _print_paper(paper):
    "Print paper"
    print(paper["title"])
    print("-"*len(paper["title"]))
    paper_copy = paper.copy()
    del paper_copy["title"]
    _print_table(paper)

def _print_table(dictionary):
    "Print dictionary as a two-column table"
    colsize = max(len(key) for key in dictionary)
    for (key, value) in dictionary.items():
        print("%s:%s%s" % (key, (colsize - len(key) + 1)*" ", value))




"This module handles writing of log messages etc."
from __future__ import print_function
from builtins import str

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-27 -- 2008-11-08"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

from sys import exit
from publish import config

def warning(message):
    "Print a warning"
    print("*** Warning: " + str(message))

def info(message):
    "Print an information message"
    print(str(message))

def error(message):
    "Print an error message"
    print("*** Error: " + str(message))

def print_summary(papers, num_found=0, num_missing=0):
    "Print summerazed result"

    print("")
    print("Summary of papers")
    print("-----------------")
    print("")

    if not (num_found == 0 and num_missing == 0):
        print("Database has %d paper(s)." % len(papers))
        print("PDF files found for %d paper(s), %d missing." % (num_found, num_missing))
        print("")

    headings = config.get("category_headings")
    categories = config.get("categories")

    # Make correct indentation for each attribute-value pair
    max_heading = max([len(headings[category]) for category in categories])

    # Count number of papers in each category
    for category in categories:
        num_papers = len([paper for paper in papers if paper["category"] == category])
        heading = headings[category]
        indentation = " " * (max_heading - len(heading))
        print("%s: %s%d" % (heading, indentation, num_papers))
    print("%s: %s%d" % ("Total", " " * (max_heading - len("Total")), len(papers)))



"This module handles filtering of papers based on attributes."
from __future__ import absolute_import

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-11 -- 2009-08-30"
__copyright__ = "Copyright (C) 2008-2009 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

from .common import short_author

def filter_papers(papers, filters):
    "Filters papers, for instance it enables to get papers from a specific year"

    filtered_papers = []
    for paper in papers:
        match = True
        for attribute in filters:
            value, should_match = filters[attribute]
            if attribute == "author" or attribute == "editor":
                if not matching_author(paper[attribute], value, should_match):
                    match = False
                    break
            else:
                for val in value.split(","):
                    if not matching_attribute(paper, attribute, val, should_match):
                        match = False
                        break
        if match:

            # Remove private key
            paper = paper.copy()
            if "private" in paper:
                del paper["private"]

            # Append paper
            filtered_papers.append(paper)

    return filtered_papers

def matching_attribute(paper, attribute, value, should_match):
    "Check if attribute matches"

    # Check for match
    match = attribute in paper and paper[attribute] == value

    # Should either match or not
    if should_match:
        return match
    else:
        return not match

def matching_author(authors, value, should_match):
    "Check if author matches"

    # Match is case-insensitive
    value = value.lower()

    # Check for match
    match = False
    for author in authors:

        # Check match against full name and abbreviation
        author_lower = author.lower()
        author_short = short_author(author).lower()
        if value in author_lower or value in author_short:
            match = True
            break

    # Should either match or not
    if should_match:
        return match
    else:
        return not match



# -*- coding: utf-8 -*-
"This module implements output for Graph ML files where authors are nodes and joint work are edges."
from future import standard_library
standard_library.install_aliases()
from builtins import str


# See https://gephi.org/users/supported-graph-formats/graphml-format/ for some more details
__author__    = "Benjamin Kehlet <benjamik@simula.no>"
__date__      = "2012-01-08 -- 2012-01-08"
__copyright__ = "Copyright (C) 2012 Benjamin Kehlet"
__license__   = "GNU GPL version 3 or any later version"

import lxml.etree as xml
import io
import itertools
from publish import config


def write(papers) :
  # Dictionary mapping author name to id and counter
  authors = {}

  # Dictionary mapping tuple of author ids to counter
  edges = {}
  for paper in papers :
    for author in paper["author"] :
      if author in authors :
        authors[author]["count"] += 1
      else : 
        authors[author] = {'id' : len(authors), 'count' : 1}


    # sort the authors to ensure the order will not cause duplicates
    authors_sorted = sorted(paper["author"])
    for author_tuple in itertools.combinations(authors_sorted, 2) :
      # look up author id and create tuple
      author_ids = tuple( [authors[single_author]["id"] for single_author in author_tuple] )

      # increment counter or add entry
      if author_ids in edges :
        edges[author_ids]  += 1 
      else :
        edges[author_ids]  = 1
  
  xml_root = xml.Element("graphml")
  
  weight_element = xml.Element("key")
  weight_element.attrib["id"] = "weight_key"
  weight_element.attrib["for"] = "edge"
  weight_element.attrib["attr.name"] = "weight"
  weight_element.attrib["attr.type"] = "double"
  xml_root.append(weight_element)

  label_element = xml.Element("key")
  label_element.attrib["id"] = "label_key"
  label_element.attrib["for"] = "node"
  label_element.attrib["attr.name"] = "label"
  label_element.attrib["attr.type"] = "string"
  xml_root.append(label_element)

  publications_element = xml.Element("key")
  publications_element.attrib["id"] = "publications_key"
  publications_element.attrib["for"] = "node"
  publications_element.attrib["attr.name"] = "publications"
  publications_element.attrib["attr.type"] = "integer"
  xml_root.append(publications_element)

  if len(config.get("mark_author")) :
    marked_element = xml.Element("key")
    marked_element.attrib["id"] = "marked_key"
    marked_element.attrib["for"] = "node"
    marked_element.attrib["attr.name"] = "marked"
    marked_element.attrib["attr.type"] = "integer"
    default_element = xml.Element("default")
    default_element.text = "0"
    marked_element.append(default_element)
    xml_root.append(marked_element)

 
  root_element = xml.Element("graph")
  root_element.attrib["edgedefault"] = "undirected"
  xml_root.append(root_element)
  

  for author, author_data in authors.items() :
    node = xml.Element("node")
    node.attrib["id"] = str(author_data["id"])
    label = xml.Element("data")
    label.attrib["key"] = "label_key"
    label.text = _filter(author)
    node.append(label)
    publications = xml.Element("data")
    publications.attrib["key"] = "publications_key"
    publications.text = str(author_data["count"])

    # Mark author if requested
    if author.strip() in config.get("mark_author") :
      marked_data = xml.Element("data")
      marked_data.attrib["key"] = "marked_key"
      marked_data.text = "1"
      node.append(marked_data)

    node.append(publications)
    root_element.append(node)

  for (i, ((source, target), count)) in enumerate(edges.items()) :
    edge = xml.Element("edge")
    edge.attrib["source"] = str(source)
    edge.attrib["target"] = str(target)
    edge.attrib["id"] = str(i)
    weight = xml.Element("data")
    weight.attrib["key"] = "weight_key"
    weight.text = str(count)
    edge.append(weight)
    root_element.append(edge)

  return xml.tostring(xml_root, encoding="utf-8", xml_declaration=True, pretty_print=True)


def _filter(s):
    "Filter string for special characters."

    s = str(s, "utf-8")


    # List of replacements
    replacements = [("--", u'-'),
                    ("$", u""),
                    ("\\mathrm", u""),
                    ('\\aa', u'å'),
                    ('\\AA', u'Å'),
                    ('\\"a', u'ä'),
                    ('\\"A', u'Ä'),
                    ('\\"o', u'ö'),
                    ('\\"O', u'Ö'),
                    ('\\o',  u'ø'),
                    ('\\O',  u'ø'),
                    ('\\ae', u'æ'),
                    ('\\AE', u'Æ'),
                    ('\\&',  u'&'),
                    # ('\\textonesuperior', "<sup>1</sup>"),
                    ('\\textendash', u'-'),
                    ('\\textemdash', u'-'),
                    ('\\textasciiacute', u'\xb4'),
                    ('\\\'a', u'\xb4'),
                    ('\\"u', u'ü'),
                    #('\\epsilon', "&epsilon;"),
                    ('\\\'\i', u'í'),
                    # ('\\textquoteright', "&apos;") # Is this correct?
                    ]

    with_brackets = [ ("{"+a+"}", b) for (a, b) in replacements]
    replacements = with_brackets+replacements


    # Iterate over replacements
    for (a, b) in replacements :
        s = s.replace(a, b)

    return s



#
# Utility for managing a database of references, typically a list of
# papers and presentations organized in different categories. The
# database is maintained in an easy-to-read and easy-to-edit format
# (emacs org-format). Using this script, lists of references can be
# filtered (for example, to include only peer-reviewed papers, only
# presentations etc) and output in one of many formats (BibTeX, HTML
# and others).

from __future__ import print_function
__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-09-24 -- 2008-12-12"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Modified by Anders Logg, 2009.

import sys

from publish.importing import import_file
from publish.exporting import export_file
from publish.validation import validate_file
from publish.printing import print_file
from publish.log import error
from publish import config

PUBLISH_VERSION = "1.0"

def extract_options(args):
    "Extract command options"

    filters = {}
    newargs = []

    for arg in args:
        if "!=" in arg:
            words = arg.split("!=")
            if not len(words) == 2:
                raise RuntimeError("Illegal option: %s" % arg)
            key, value = words
            filters[key] = (value, False)
        elif "=" in arg:
            words = arg.split("=")
            if not len(words) == 2:
                raise RuntimeError("Illegal option: %s" % arg)
            key, value = words
            if config.has_key(key) :
                config.set_from_string(key, value)
            else:
                filters[key] = (value, True)
        else:
            newargs.append(arg)

    return newargs, filters

def print_usage():
    "Print help text"
    print("""\
Usage: publish <command> [command/global option]... [filename]

Commands: import, validate, export, print

For more information, refer to the Publish man page ('man publish')
or the Publish User Manual.
""")
    print_copyright()

def print_copyright():
    "Print copyright notice"
    print("""\
Copyright (C) 2008 Anna Logg (Logg Systems).
Licensed under the GNU GPL version 3 or any later version.""")

def print_version():
    "Print version number"
    print("This is Publish, a distributed publication management system, version %s." % PUBLISH_VERSION)
    print("")
    print_copyright()

def run_import(args, filters):
    "Run import command"
    for filename in args:
        import_file(filename, filters)

def run_export(args, filters):
    "Run export command"
    for filename in args:
        export_file(filename, filters)

def run_validate(args, filters):
    "Run validate command"
    if len(args) > 0:
        for filename in args:
            validate_file(filename)
    else:
        validate_file(None)

def run_print(args, filters):
    "Run print command"
    print_file(filters)

# Available commands
commands = {"import":   run_import,
            "export":   run_export,
            "validate": run_validate,
            "print":    run_print}

def main(args):

    # Print help text if requested
    if len(args) == 1 and args[0] in ("-h", "--help"):
        print_usage()
        return 0

    # Print version number if requested
    if len(args) == 1 and args[0] in ("-v", "--version"):
        print_version()
        return 0

    # Check command
    if len(args) >= 1 and args[0] in commands:

        # Choose command
        command = commands[args[0]]

        # Extract options
        args, filters = extract_options(args[1:])

        # Run command if running in debug mode
        if config.get("debug"):
            return command(args, filters)

        # Run command and catch exception if not in debug mode
        if 1:
            return command(args, filters)
        try:
            return command(args, filters)
        except Exception as message:
            error(message)
            return 2
    # Illegal arguments
    print_usage()
    return 2

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



"""
Default configuration data for publish.

User can make a publish_config.py file with

from publish.config.default.py import *
"""

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-28 -- 2008-11-03"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

MARKER_FOR_IMPORT_0123456789 = 'variable for marking import of this module'

# Modified by Anders Logg 2011
# Modified by Benjamin Kehlet 2012-2013
# Last modified: 2013-04-23

database_filename       = "papers.pub"
local_venues_filename   = "venues.list"
authornames_filename    = "authors.list"
invalid_filename_prefix = "invalid_papers"

matching_distance_strong = 0.1
matching_distance_weak   = 0.5

autofix  = False
debug    = False

pdf_viewer = "evince"
view_pdf = True
pdf_dir = "papers"

headline = ""
repeat_headline = False

compact = False

require_page_range = False
page_separator = "-"

# Set a prefix for the class names when generating html.
# Example: class="publish_item_title"
html_class_prefix = "publish"
html_add_internal_links = False

# Should numbering in report be global
# or start over for each category
# NOTE: Not fully implemented for all report types.
global_numbering = False

# Control whether output should contain categories or just be a list
# NOTE: Not fully implemented for all report types.
skip_categories = False

# Control whether LaTeX bibitems should use labels [J1], [J2] etc
# NOTE: Not fully implemented for all report types.
use_labels = False

# When searching for duplicates, handle the case where one publication
# is a talk and the other is not and a non-duplicate
# Usefull since we often have talks with the same title as a paper
talks_dont_duplicate = True

# When importing from bibtex, check that categories are according to
# the bibtex standard.
# If False, then all categories allowed.
use_standard_categories = True

# When importing a file, do not validate and do not process duplicates
simple_import = False

# Common typos
typos_common = {"\\a\\'": "\\'",
                "": " ",
                "{\\r A}": "{\\AA}",
                "{\\r a}": "{\\aa}"}

# Typos in author strings
typos_author = {"&": None,
                "(": None,
                ")": None,
                ",": None}

# Typos in editor strings
typos_editor = typos_author

# Typos in key strings
typos_key = {"\_": ""}

# Collect typos
typos = {"common": typos_common,
         "author": typos_author,
         "editor": typos_editor,
         "key":    typos_key}

institutions = ("Carleton University",
                "Finite Element Center",
                "InterMedia, University of Oslo",
                "Norwegian University of Science and Technology",
                "Numerical Objects A.S.",
                "University of Castilla-La Mancha",
                "Simula Research Laboratory")

schools = ("Department of Informatics, University of Oslo",
           "Faculty of Mathematics and Natural Sciences, University of Oslo")

publishers = ("Birkh{\\\"a}user",
              "Cambridge University Press",
              "Chapman and Hall",
              "Computers in Cardiology",
              "Elsevier Science",
              "EuroPACS",
              "European Association of Geoscientists \\& Engineers",
              "European Community on Computational Methods in Applied Sciences",
              "Informing Science Press",
              "John von Neumann Institute for Computing",
              "Kluwer Academic Publishers",
              "Literat{\\\"u}r Yayincilik Ltd.",
              "MGNet",
              "Nationellt centrum f{\\\"o}r matematikutbildning, NCM",
              "North Holland",
              "Nova Science Publishers",
              "SIAM",
              "SPIE",
              "Springer",
              "Swansea",
              "Tapir Academic Press",
              "Taylor \\& Francis Books Ltd",
              "Wiley Press")

# Words that should not be capitalized
lowercase = ["a", "an", "and", "at", "for", "in", "of", "on", "the", "to", "by"]

# Words that should be capitalized
uppercase = {
    "python": "{P}ython",
    "cython": "{C}ython",
    "fortran": "{FORTRAN}",
    "c++":    "{C}++",
    "matlab": "{MATLAB}",
    "xml": "{XML}",
    "java": "{Java}",
    "perl": "{P}erl",
    "ruby": "{R}uby}",
    "diffpack": "{D}iffpack",
    "dolfin": "{DOLFIN}",
    "swig": "{SWIG}",
    "instant": "{I}nstant",
    "oo": "{OO}",
    "planck": "{Planck}",
    "fouirer": "{F}ourier",
    "fokker-planck": "{Fokker-Planck}",
    "navier-stokes": "{Navier-Stokes}",
    "fenics": "{FEniCS}",
    "newton": "{N}ewton",
    "markov":"{Markov}",
    "boltzmann": "{B}oltzmann",
    "lattice-boltzmann": "{Lattice-Boltzmann}",
    "lagrange":"{Lagrange}",
    "laplacian":"{Laplacian}",
    "laplace": "{L}aplace",
    "poisson": "{P}oisson",
    "poisson's": "{P}oisson's",
    "ode": "{ODE}",
    "pde": "{PDE}",
    "fdm": "{FDM}",
    "fvm": "{FVM}",
    "fem": "{FEM}",
    "fsi": "{FSI}",
    " fe ": " {FE} ",
    " fd ": " {FD} ",
    "pdes": "{PDE}s",
    "odes": "{ODE}s",
    "ode's": "{ODE}'s",
    "pde's": "{PDE}'s",
    "infiniband": "{I}nfini{B}and",
    "ethernet": "{E}thernet",
    "mapreduce": "{M}ap{R}educe",
    "tcp": "{TCP}",
    "cpu": "{CPU}",
    "cpus": "{CPU}s",
    "gpu": "{GPU}",
    "gpus": "{GPU}s",
    "mpi": "{MPI}",
    "openmp": "{O}pen{MP}",
    "cuda": "{CUDA}",
    "cfd": "{CFD}",
    "hpc": "{HPC}",
    "mpeg":"{MPEG}",
    }

journals = [("none", "none", "xxxx-xxxx"),
            ("ACM Trans. Math. Software", "ACM Transactions on Mathematical Software", "0098-3500"),
            ("ACM Comput. Surv.", "ACM Computing Surveys", "0360-0300"),
            ("ACM SIGSOFT Software Engineering Notes", "ACM SIGSOFT Software Engineering Notes", "01635848"),
            ("ACM Trans. Multimed. Comput. Comm. Appl.", "ACM Transactions on Multimedia Computing, Communications, and Applications", "1551-6857"),
            ("ACM Trans. Software Eng. Meth.", "ACM Transactions on Software Engineering and Methodology", "1049-331X"),
            ("Adv. Comput. Math.", "Advances in Computational Mathematics", "1019-7168"),
            ("Adv. Water Resour.", "Advances in Water Resources", "0309-1708"),
            ("Ann. Biomed. Eng.", "Annals of Biomedical Engineering", "0090-6964"),
            ("Appl. Cognit. Psychol.", "Applied Cognitive Psychology", "0888-4080"),
            ("Appl. Math. Comput.", "Applied Mathematics and Computation", "0096-3003"),
            ("Appl. Math. Finance", "Applied Mathematical Finance", "1350-486X"),
            ("Appl. Math. Model.", "Applied Mathematical Modelling", "0307-904X"),
            ("Appl. Mech. Eng.", "Applied Mechanics and Engineering", "1425-1655"),
            ("Arch. Comput. Methods Eng.", "Archives of Computational Methods in Engineering", "1134-3060"),
            ("Behaviour \\& Information Technology", "Behaviour \\& Information Technology", "0144-929X"),
            ("Biophys. J.", "Biophysical Journal", "0006-3495"),
            ("BIT", "BIT Numerical Mathematics", "0006-3835"),
            ("Camput-Wide Information Systems", "Campus-Wide Information Systems", "1065-0741"),
            ("Comput. Biol. Med.", "Computers in Biology and Medicine", "0010-4825"),
            ("Comput. Comm.", "Computer Communications", "0140-3664"),
            ("Comput. Geosci.", "Computational Geosciences", "1420-0597"),
            ("Comput. Phys.", "Communications in Computational Physics", "1814-2406"),
            ("Comput. Math. Appl.", "Computers \\& Mathematics with Applications", "0898-1221"),
            ("Comput. Mech.", "Computational Mechanics. Solids, Fluids, Engineered Material, Aging Infrastructure, Molecular Dynamics, Heat Transfer, Manufacturing Processes, Optimization, Fracture \& Integrity", "0178-7675"),
            ("Comput. Methods Appl. Mech. Engrg", "Computer Methods in Applied Mechanics and Engineering", "0045-7825"),
            ("Comput. Meth. Biomech. Biomed. Eng.", "Computer Methods in Biomechanics and Biomedical Engineering", "1025-5842"),
            ("Comput. Sci. Eng.", "Computing in Science \\& Engineering", "1521-9615"),
            ("Comput. Vis. Sci.", "Computing and Visualization in Science", "1432-9360"),
            ("Continent. Shelf Res.", "Continental Shelf Research", "0278-4343"),
            ("Empir. Software Eng.", "Empirical Software Engineering", "1382-3256"),
            ("Eng. Anal. Bound. Elem.", "Engineering Analysis with Boundary Elements", "0955-7997"),
            ("Formal Aspect. Comput.", "Formal Aspects of Computing", "0934-5043"),
            ("Future Generat. Comput. Syst.", "Future Generation Computer Systems", "0167-739X"),
            ("Global J. Pure Appl. Math.", "Global Journal of Pure and Applied Mathematics", "0973-1768"),
            ("IEEE/ACM Transactions on Networking", "IEEE/ACM Transactions on Networking", "1063-6692"),
            ("IEEE Comm. Lett.", "IEEE Communications Letters", "1089-7798"),
            ("IEEE Comm. Mag.", "IEEE Communications Magazine", "0163-6804"),
            ("IEEE Comput. Appl. Power Mag.", "IEEE Computer Applications in Power Magazine", "0895-0156"),
            ("IEEE Comput. Architect. Lett.", "IEEE Computer Architecture Letters", "1556-6056"),
            ("IEEE Control Syst. Mag.", "IEEE Control Systems Magazine", "1066-033X"),
            ("IEEE Distr. Syst. Online", "IEEE Distributed Systems Online", "1541-4922"),
            ("IEEE Micro Magazine", "IEEE Micro Magazine", "0272-1732"),
            ("IEEE Netw. Mag. Global Inform. Exchange", "IEEE Network: The Magazine of Global Information Exchange", "0890-8044"),
            ("IEEE Software Mag.", "IEEE Software Magazine", "0740-7459"),
            ("IEEE Trans. Biomed. Eng.", "IEEE Transactions on Biomedical Engineering", "0018-9294"),
            ("IEEE Trans. Comput.", "IEEE Transactions on Computers", "0018-9340"),
            ("IEEE Transactions on Industrial Informatics", "IEEE Transactions on Industrial Informatics", "1551-3203"),
            ("IEEE Trans. Parallel Distr. Syst.", "IEEE Transactions on Parallel and Distributed Systems", "1045-9219"),
            ("IEEE Trans. Software Eng.", "IEEE Transactions on Software Engineering", "0098-5589"),
            ("IEEE Trans. Wireless Comm.", "IEEE Transactions on Wireless Communications", "1536-1276"),
            ("IET Communications", "IET Communications", "1751-8628"),
            ("IMA J. Appl. Math.", "IMA Journal of Applied Mathematics", "0272-4960"),
            ("IMA J. Numer. Anal.", "IMA Journal of Numerical Analysis", "0272-4979"),
            ("Inform. Process. Lett.", "Information Processing Letters", "0020-0190"),
            ("Inform. Software Tech.", "Information and Software Technology", "0950-5849"),
            ("ISA Transactions", "ISA Transactions", "0019-0578"),
            ("Appl. Mech. Eng.", "International Journal of Applied Mechanics and Engineering", "1425-1655"),
            ("J. Comput. Methods Eng. Sci. Mech.", "International Journal for Computational Methods in Engineering Science and Mechanics", "1550-2287"),
            ("International Journal of Autonomous and Adaptive Communications Systems", "International Journal of Autonomous and Adaptive Communications Systems", "1754-8632"),
            ("International Journal of Bioelectromagnetism", "International Journal of Bioelectromagnetism", "1456-7857"),
            ("Int. J. Comput. Eng. Sci.", "International Journal of Computational Engineering Science", "1465-8763"),
            ("IJCSE", "International Journal of Computational Science and Engineering", "1742-7185"),
            ("Int. J. Comput. Math.", "International Journal of Computer Mathematics", "0020-7160"),
            ("Int. J. Forecast", "International Journal of Forecasting", "0169-2070"),
            ("Int. J. Fluid Mech. Res.", "International Journal of Fluid Mechanics Research", "1064-2277"),
            ("Int. J. Numer. Anal. Model.", "International Journal of Numerical Analysis and Modeling", "1705-5105"),
            ("Int. J. Nonlinear Sci. Numer. Simul.", "International Journal of Nonlinear Sciences and Numerical Simulation", "1565-1339"),
            ("Int. J. Nonlinear Model. Sci. Eng.", "International Journal of Nonlinear Modelling in Science and Engineering", "1472-085X"),
            ("Int. J. Numer. Meth. Eng.", "International Journal for Numerical Methods in Engineering", "0029-5981"),
            ("Int. J. Numer. Meth. Fluid.", "International Journal for Numerical Methods in Fluids", "0271-2091"),
            ("Int. J. Proj. Manag.", "International Journal of Project Management", "0263-7863"),
            ("International Journal of Pure and Applied Mathematics", "International Journal of Pure and Applied Mathematics", "1311-8080"),
            ("Int. J. Software Eng. Knowl. Eng.", "International Journal of Software Engineering and Knowledge Engineering", "0218-1940"),
            ("J. Comput. Acoust", "Journal of Computational Acoustics", "0218-396X"),
            ("J. Comput. Appl. Math.", "Journal of Computational and Applied Mathematics", "0377-0427"),
            ("J. Comput. Finance", "Journal of Computational Finance", "1460-1559"),
            ("J. Comput. Neurosci.", "Journal of Computational Neuroscience", "0929-5313"),
            ("J. Comput. Phys.", "Journal of Computational Physics", "0021-9991"),
            ("Journal d'Analyse Mathematique", "Journal d'Analyse Mathematique", "0021-7670"),
            ("J. Geodyn.", "Journal of Geodynamics", "0264-3707"),
            ("J. Geophys. Res. B", "Journal of Geophysical Research (JGR)-B", "0196-6936"),
            ("Journal of Medical Ultrasonics", "Journal of Medical Ultrasonics", "1346-4523"),
            ("J. Netw. Comput. Appl.", "Journal of Network and Computer Applications", "1084-8045"),
            ("J. Numer. Math.", "Journal of Numerical Mathematics", "1570-2820"),
            ("J. Object Tech.", "Journal of Object Technology", "1660-1769"),
            ("J. Sci. Comput.", "Journal of Scientific Computing", "0885-7474"),
            ("J. Software Mainten. Evol. Res. Pract.", "Journal of Software Maintenance and Evolution: Research and Practice", "1532-060X"),
            ("J. Stat. Software", "Journal of Statistical Software", "1548-7660"),
            ("J. Syst. Architect.", "Journal of Systems Architecture", "1383-7621"),
            ("J. Syst. Software", "Journal of Systems and Software", "0164-1212"),
            ("Math. Biosci.", "Mathematical Biosciences", "0025-5564"),
            ("Math. Comp.", "Mathematics of Computation", "0025-5718"),
            ("Math. Comput. Simulat.", "Mathematics and Computers in Simulation", "0378-4754"),
            ("Math. Geol.", "Mathematical Geology", "0882-8121"),
            ("Metheoritics Planet. Sci.", "Metheoritics \\& Planetary Science", "1086-9379"),
            ("Mobile Network. Appl.", "Mobile Networks and Applications", "1383-469X"),
            ("Model. Identif. Control", "Modeling, Identification and Control. A Norwegian Research Bulletin", "0332-7353"),
            ("Multi Agent and Grid Systems", "Multi Agent and Grid Systems", "1574-1702"),
            ("Nord. J. Comput.", "Nordic Journal of Computing", "1236-6064"),
            ("Norwegian Journal of Geography/ Norsk Geografisk Tidsskrift", "Norwegian Journal of Geography/ Norsk Geografisk Tidsskrift", "0029-1951"),
            ("Numer. Linear Algebra Appl.", "Numerical Linear Algebra with Applications", "1070-5325"),
            ("Numer. Methods Partial Differential Equations", "Numerical Methods for Partial Differential Equations. An International Journal", "0749-159X"),
            ("Numer. Math.", "Numerische Mathematik", "0029-599X"),
            ("Otic. Comm.", "Optics Communications", "0030-4018"),
            ("Perform. Eval.", "Performance Evaluation", "0166-5316"),
            ("Personal and Ubiquitous Computing", "Personal and Ubiquitous Computing", "1617-4909"),
            ("Phys. Earth Planet. In.", "Physics of the Earth and Planetary Interiors", "0031-9201"),
            ("Phys. Rev. B Condens. Matter", "Physical Review B: Condensed Matter and Materials Physics", "1098-0121"),
            ("Proc. IEEE Comput. Soc. Bioinformatics Conf. IEEE Comput. Soc. Bioinformatic", "Proceedings / IEEE Computer Society Bioinformatics Conference. IEEE Computer Society Bioinformatics Conference", "1555-3930"),
            ("Research Letters in Communications", "Research Letters in Communications", "1687-6741"),
            ("Scand. J. Inform. Syst.", "Scandinavian Journal of Information Systems", "0905-0167"),
            ("Sci. Program.", "Scientific Programming", "1058-9244"),
            ("SIAM J. Numer. Anal.", "SIAM Journal on Numerical Analysis", "0036-1429"),
            ("SIAM J. Sci. Comput.", "SIAM Journal on Scientific Computing", "1064-8275"),
            ("Software Process Improv. Pract.", "Software Process: Improvement and Practice", "1077-4866"),
            ("Software Syst. Model.", "Software and Systems Modeling", "1619-1366"),
            ("Stroke", "Stroke", "0039-2499"),
            ("Transactions on Machine Learning and Data Mining", "Transactions on Machine Learning and Data Mining", "1865-6781"),
            ("Transp. Porous Media", "Transport in Porous Media", "0169-3913"),
            ("Wireless Pers. Comm.", "Wireless Personal Communications", "0929-6212")]

# Categories
categories = ("articles",
              "books",
              "edited",
              "chapters",
              "refproceedings",
              "proceedings",
              "reports",
              "manuals",
              "theses",
              "courses",
              "talks",
              "posters",
              "publicoutreach",
              "preprint",
              "misc")

# Headings for categories
category_headings = {"articles":       "Articles in International Journals",
                     "books":          "Books",
                     "edited":         "Edited Books",
                     "chapters":       "Chapters in Books",
                     "refproceedings": "Refereed Proceedings",
                     "proceedings":    "Conference Proceedings",
                     "reports":        "Technical Reports",
                     "manuals":        "Manuals",
                     "theses":         "Theses",
                     "courses":        "Courses",
                     "talks":          "Talks",
                     "posters":        "Posters",
                     "publicoutreach": "Public Outreach",
                     "preprint":       "Preprints",
                     "misc":           "Other Publications"}

# Labels for categories
category_labels = {"articles":       "J",
                   "books":          "B",
                   "edited":         "E",
                   "chapters":       "C",
                   "refproceedings": "P",
                   "proceedings":    "Pc",
                   "reports":        "R",
                   "manuals":        "M",
                   "theses":         "T",
                   "courses":        "Co",
                   "talks":          "Ta",
                   "posters":        "Po",
                   "publicoutreach": "Or",
                   "preprint":       "Pr",
                   "misc":           "Mi"}

# Attributes for categories (tuple means at least one of listed attributes required)
category_attributes = {"articles":       ("author", "title", "journal", "year", "status"),
                       "books":          ("author", "title", "publisher", "year", "status"),
                       "edited":         ("author", "title", "publisher", "year", "status"),
                       "chapters":       ("author", "title", "editor", "publisher", "year", "status"),
                       "refproceedings": ("author", "title", "booktitle", "year", "status"),
                       "proceedings":    ("author", "title", "booktitle", "year", "status"),
                       "reports":        ("author", "title", "institution", "year", "status"),
                       "manuals":        ("author", "title", "status"),
                       "theses":         ("author", "title", "school", "year", "thesistype", "status"),
                       "courses":        ("author", "title", "code", "institution", "year", "status"),
                       "talks":          ("author", "title", "meeting", "year", "status"),
                       "posters":        ("author", "title", "meeting", "year", "status"),
                       "publicoutreach": ("author", "title", "meeting", "year", "status"),
                       "preprint":       ("author", "title", "year", "status"),
                       "misc":           ("title", "status")}

# Venues for categories (will be matched against a list of allowed values)
category_venues = {"articles":       "journal",
                   "books":          "publisher",
                   "edited":         "publisher",
                   "chapters":       "publisher",
                   "refproceedings": None,
                   "proceedings":    None,
                   "reports":        "institution",
                   "manuals":        None,
                   "theses":         "school",
                   "courses":        "institution",
                   "talks":          "meeting",
                   "posters":        "meeting",
                   "publicoutreach": "meeting",
                   "preprint":       None,
                   "misc":           None}

# List of ordered attributes (so we rewrite pub files in a fixed order)
ordered_attributes = ("title",
                      "key",
                      "author", "editor",
                      "year",
                      "date",
                      "journal", "booktitle", "institution", "meeting",
                      "publisher",
                      "volume",
                      "number",
                      "chapter",
                      "pages",
                      "num_pages",
                      "doi",
                      "arxiv",
                      "pdf",
                      "url",
                      "status",
                      "duplicate",
                      "allowed_duplicates",
                      "tags",
                      "private",
                      "fixme")

# Entry types (BibTeX)
entrytypes = ("article",
              "book",
              "booklet",
              "conference",
              "inbook",
              "incollection",
              "inproceedings",
              "manual",
              "mastersthesis",
              "misc",
              "phdthesis",
              "proceedings",
              "techreport",
              "preprint",
              "unpublished")

# Attributes for entry types (tuple means at least one of listed attributes required)
entrytype_attributes = {"article":      ("author", "title", "journal", "year"),
                        "book":         (("author", "editor"), "title", "publisher", "year"),
                        "booklet":      ("title",),
                        "conference":   ("author", "title", "booktitle", "year"),
                        "inbook":       (("author", "editor"), "title", ("chapter", "pages"), "publisher", "year"),
                        "incollection": ("author", "title", "booktitle", "publisher", "year"),
                        "inproceedings":("author", "title", "booktitle", "year"),
                        "manual":       ("title",),
                        "mastersthesis": ("author", "title", "school", "year"),
                        "misc":         (),
                        "phdthesis":    ("author", "title", "school", "year"),
                        "proceedings":  ("title", "year"),
                        "techreport":   ("author", "title", "institution", "year"),
                        "preprint":     ("author", "title"),
                        "unpublished":  ("author", "title", "note")}

# Mapping from entry type to category
entrytype2category = {"article":          "articles",
                      "book":             None,            # special case: books or edited
                      "book proceedings": "edited",
                      "inbook":           "chapters",
                      "inproceedings":    "proceedings",
                      "conference":       "proceedings",
                      "techreport":       "reports",
                      "manual":           "manuals",
                      "preprint":         "preprint",
                      "phdthesis":        None,            # special case: theses
                      "mastersthesis":    None,            # special case: theses
                      "misc":             None}            # special case: misc, talks, or courses

# Mapping from category to entry type
category2entrytype = {"articles":       "article",
                      "books":          "book",
                      "edited":         "book",
                      "chapters":       "inbook",
                      "refproceedings": "inproceedings",
                      "proceedings":    "inproceedings",
                      "reports":        "techreport",
                      "manuals":        "manual",
                      "theses":         None,
                      "courses":        "misc",
                      "talks":          "misc",
                      "preprint":       "preprint",
                      "misc":           "misc"}

# Thesis type strings
thesistype_strings = {"phd":     "Ph.D. Thesis",
                      "msc":     "M.Sc. Thesis",
                      "lic":     "Lic. Thesis",
                      "diploma": "Dipl. Thesis"}



"This module handles interaction between the system and the user."
from __future__ import print_function
from builtins import input
from builtins import str
from builtins import range

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-27 -- 2008-11-12"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

from publish import config

def ask_user_yesno(question, default="yes"):
    "Ask a yes-no question"

    if default == "yes":
        options = "[y]/n"
        default_return = True
    else:
        options = "y/[n]"
        default_return = False
    
    if config.get("autofix"):
        print(question + (" (%s):" % options))
        print("  Autofix enabled, choosing default (%s)." % default)
        return default_return
    
    while True:
        answer = input(question + (" (%s): " % options))
        if answer == "y":
            return True
        elif answer == "n":
            return False
        elif answer == "":
            return default_return
        print('Please answer "y" or "n".')

def ask_user_alternatives(question, alternatives):
    "Ask for an option"

    while True:
        print(question)
        n = len(alternatives)
        for i in range(n):
            alternative = alternatives[i]
            print("  [%d] %s" % (i+1, alternative))
        numbers = ", ".join([str(i+1) for i in range(n-1)]) + " or " + str(n)
        if config.get("autofix"):
            print("  Autofix enabled, choosing default (1).")
            return 0

        s = input("Please enter %s (or press return to choose [1]): " % numbers)
        if s.strip() == "":
            choice = 1
        else:
            try :
                choice = int(s)
            except ValueError :
                # If answer could not be converted to int, set to illegal int value
                choice = -1

        if (choice - 1) in range(n):
            return choice - 1

        print("Illegal option.")



"This module implements validation of data from different file formats."
from __future__ import print_function
from builtins import input
from builtins import str
from builtins import range

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-27 -- 2008-12-12"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Last changed: 2012-08-31
# Modified by Benjamin Kehlet 2012

import os

from publish.common import is_valid, pstr
from publish.formats import pub
from publish.log import warning, print_summary
from publish import config
from publish.algorithms import distance
from publish.interaction import ask_user_yesno, ask_user_alternatives
from publish.database import read_database, save_database, save_invalid_papers
from publish.keygeneration import generate_keys
from publish.merging import process_duplicates

# Maps complete journal name with its issn-number
long2issn = config.get("long2issn")
long2issn = {}

# Maps complete journal name with its abbreviation
long2short = config.get("long2short")
long2short = {}

# Maps journal abbreviation with its complete name
short2long = config.get("short2long")
short2long = {}

def validate_file(filename=None):
    "Validate data in file"

    # Use default database if file is not specified
    if filename is None and not os.path.isfile(config.get("database_filename")):
        print("No file specified and no database found, nothing to do.")
        return

    # Open and read database
    papers = read_database(filename)

    # Validate papers
    (database_papers, invalid_papers) = validate_papers(papers)

    # Generate keys
    database_papers = generate_keys(papers)

    # Check for PDF files
    (num_found, num_missing) = check_pdf_files(papers)

    # Checking for duplicates
    database_papers = process_duplicates(database_papers)

    # Print summary
    print_summary(papers, num_found, num_missing)

    # Save papers to database
    save_database(database_papers)
    save_invalid_papers(invalid_papers)

def validate_papers(papers):
    "Validate list of papers"

    if len(papers) == 0:
        return ([], [])

    print("")
    print("Validating papers")
    print("-----------------")
    print("")

    # Validate each paper in the list of papers
    for paper in papers:
        _validate_paper(paper)

    # Extract valid and invalid papers
    valid_papers   = [paper for paper in papers if is_valid(paper)]
    invalid_papers = [paper for paper in papers if not is_valid(paper)]

    # Print a short summary
    print("")
    print("Validated %d paper(s) ok." % len(valid_papers))
    print("Found %d invalid paper(s)." % len(invalid_papers))
    print("")

    return (valid_papers, invalid_papers)

def _validate_paper(paper):
    "Validate paper, marking it as invalid of not valid"

    print("Validating paper: %s" % pstr(paper))

    # Validate status
    if is_valid(paper):
        _validate_paper_status(paper)

    # Validate that no categories are missing
    if is_valid(paper):
        _validate_paper_categories(paper)

    # Validate all paper strings for typos
    if is_valid(paper):
        _validate_paper_typos(paper)

    # Validate venue
    if is_valid(paper):
        _validate_paper_venue(paper)

    if is_valid(paper) :
        _validate_paper_authors(paper)

    # Validate title
    if is_valid(paper):
        _validate_paper_title(paper)

    # Validate page range
    if is_valid(paper):
        _validate_paper_pages(paper)

def _validate_paper_status(paper):
    "Validate that status is specified"

    # Set status if missing
    if not "status" in paper:
        print('  Status is not defined, assuming status is "published".')
        paper["status"] = "published"

def _validate_paper_categories(paper):
    "Validate that no attributes are missing"

    # Check that category is specified
    if not "category" in paper:
        raise RuntimeError("Unable to validate paper, unknown category.")

    # Check that the paper holds all required attributes
    category = paper["category"]
    category_attributes = config.get("category_attributes")
    for attribute in category_attributes[category]:

        if isinstance(attribute, tuple):
            if not len([a for a in attribute if a in paper]) > 0:
                paper["invalid"] = True
                missing = str(attribute)
                break
        else:
            if not attribute in paper:
                paper["invalid"] = True
                missing = str(attribute)
                break

    if not is_valid(paper):
        print('  Skipping paper (missing attribute "%s")' % missing)
        if not config.get("autofix"):
            input("  Press return to continue.")

def _validate_paper_venue(paper):
    "Validate that the venue (journal, conference etc) is correct"

    # Get venue type
    category = paper["category"]
    category_venues = config.get("category_venues")
    venue_type = category_venues[category] # "journal", "booktitle", etc

    # Skip if venue is None (nothing to check)
    if venue_type is None:
        return

    # Get list of known venues
    known_venues = config.get(venue_type + "s")

    # Check that venue is valid
    venue_name = paper[venue_type]
    if not venue_name in known_venues:
        print("")
        print('  Unknown %s: "%s"' % (venue_type, venue_name))
        suggested_venue = _suggest_venue(venue_name, known_venues)
        if suggested_venue is None:
            if ask_user_yesno('  Would you like to add %s "%s"?' % (venue_type, venue_name), "no"):
                _add_venue(venue_type, venue_name)
            else:
                print("  Skipping paper.")
                paper["invalid"] = True
        else:
            print('  Suggested %s: "%s"' % (venue_type, suggested_venue))
            alternative = ask_user_alternatives("  Unknown %s, what should I do?" % venue_type,
                                                ("Replace %s." % venue_type,
                                                 "Add %s." % venue_type,
                                                 "Skip paper."))
            print("")
            if alternative == 0:
                paper[venue_type] = suggested_venue
            elif alternative == 1:
                _add_venue(venue_type, venue_name)
            elif alternative == 2:
                print("  Skipping paper (unable to guess the right %s)" % venue_type)
                input("  Press return to continue.")
                paper["invalid"] = True
            else:
                raise RuntimeError("Unknown option.")

def _validate_paper_authors(paper) :
    "Validate spelling of the author names"

    allowed_author_names = config.get("allowed_author_names")

    # Skip if there is nothing to check
    if allowed_author_names is None:
        return

    for i, author in enumerate(paper["author"]) :
        if not author in allowed_author_names :
            print("\n  Unknown author: \"%s\"" % author)
            suggested_author = _suggest_venue(author, allowed_author_names)
            if suggested_author is None:
                if ask_user_yesno('  Would you like to add %s"?' % (author), "no"):
                    _add_author(author)
                else:
                    print("  Skipping paper.")
                    paper["invalid"] = True
            else:
                print('  Suggested "%s"' % (suggested_author))
                alternative = ask_user_alternatives("  Unknown author, what should I do?",
                                                    ("Replace author.",
                                                     "Add author.",
                                                     "Skip paper."))
                print("")
                if alternative == 0:
                    paper["author"][i] = suggested_author
                elif alternative == 1:
                    _add_author(suggested_author)
                elif alternative == 2:
                    print("  Skipping paper (unable to guess the right author)")
                    input("  Press return to continue.")
                    paper["invalid"] = True
                else:
                    raise RuntimeError("Unknown option.")



def _validate_paper_title(paper):
    "Validate that the title is correct, fix capitalization"

    # Fix capitalization
    title = paper["title"]
    for separator in (" ", "-"):

        words = title.split(separator)

        lowercase = config.get("lowercase")
        uppercase = config.get("uppercase")
        new_words = []
        for i in range(len(words)):
            word = words[i]
            if word == "":
                continue
            if word.lower() in lowercase:
                word = word.lower()
            elif word.lower() in uppercase:
                word = uppercase[word.lower()]
            else:
                word = word[0].upper() + word[1:]
            new_words.append(word)

        title = separator.join(new_words)

    paper["title"] = title

def _validate_paper_pages(paper):
    "Validate page range"

    # Only check if we have pages
    if not "pages" in paper:
        return
    pages = paper["pages"]

    invalid = False
    new_pages = pages

    # Check if page page must contain "-"
    if config.get("require_page_range") and not "-" in pages:
        invalid = True
    if "-" in pages:
        if "--" in pages:
            first, last = pages.split("--")[:2]
        else:
            first, last = pages.split("-")[:2]
            if len(first) == 0 or len(last) == 0:
                invalid = True

            # Reformat string
            new_pages = first.strip() + config.get("page_separator") + last.strip()

    # Check for invalid page string
    if invalid:
        paper["invalid"] = True
        print("  Incorrectly formatted page string: " + pages)
        if not config.get("autofix"):
            input("  Skipping paper, press return to continue.")
        else:
            print("  Skipping paper.")
        return


    # Check if string was changed
    if not new_pages == pages:
        print("  Incorrectly formatted page string: " + pages)
        print("  Suggested correction:              " + new_pages)
        if ask_user_yesno("  Would you like to accept the suggested correction:"):
            print("  Correcting page string.")
            paper["pages"] = new_pages
        else:
            paper["invalid"] = True
            if not config.get("autofix"):
                input("  Skipping paper, press return to continue.")
            else:
                print("  Skipping paper.")

def _validate_paper_typos(paper):
    "Validate all paper strings for typos"

    typos = config.get("typos")

    # Check all attributes
    for attribute in paper:

        # Extract typos to check
        attribute_typos = typos["common"].copy()
        if attribute in typos:
            for typo in typos[attribute]:
                attribute_typos[typo] = typos[attribute][typo]

        # Get attribute value and convert to tuple
        value = paper[attribute]
        if isinstance(value, tuple):
            value_tuple = value
        else:
            value_tuple = (value,)

        # Check all values in tuple
        new_values = []
        for value in value_tuple:
            for typo in attribute_typos:
                replacement = attribute_typos[typo]

                # Check for typo
                if typo in value:
                    print("  Incorrectly formatted %s string: %s" % (attribute, str(value)))

                    if replacement is None:

                        # Found no replacement, skip paper
                        paper["invalid"] = True
                        if config.get("autofix"):
                            print("  Skipping paper")
                        else:
                            input("  Skipping paper, press return to continue.")
                        return

                    else:

                        # Found replacement
                        value = value.replace(typo, replacement)
                        print('  Replacing typo "%s" with "%s".' % (typo, replacement))
                        if not config.get("autofix"):
                            print("  Press return to continue.")

            new_values.append(value)

        # Assign corrected value
        if isinstance(paper[attribute], tuple):
            paper[attribute] = tuple(new_values)
        else:
            paper[attribute] = new_values[0]

def _suggest_venue(venue, known_venues):
    "Suggest venue name"

    matching_distance = config.get("matching_distance_weak")
    suggested_venue = None
    min_distance = 1.0

    # Look for closes match
    for v in known_venues:
        d = distance(venue.lower(), v.lower())
        if d < matching_distance and d < min_distance:
            suggested_venue = v
            min_distance = d

    # If not found try matching substrings
    if suggested_venue is None:
        for v in known_venues:
            d = distance(venue.lower(), v.lower())
            if venue in v or v in venue and d < min_distance:
                suggested_venue = v

    return suggested_venue

def _add_venue(venue_type, venue_name):
    "Add venue to known venues"

    # Append to list of known venues (remember at run-time)
    known_venues = config.get(venue_type + "s")
    known_venues.append(venue_name)

    # Append to file
    filename = config.get("local_venues_filename")
    try:
        file = open(filename, "a")
        file.write("%s: %s\n" % (venue_type, venue_name))
    except:
        raise RuntimeError('Unable to add local venue to file "%s".' % filename)

def _add_author(author_name) :
    allowed_author_names = config.get("allowed_author_names")
    allowed_author_names.add(author_name)

    # Append to file
    filename = config.get("authornames_filename")

    try:
        file = open(filename, "a")
        file.write(author_name.strip()+"\n")
    except:
        raise RuntimeError('Unable to author to file: "%s".' % filename)


def check_pdf_files(papers):
    "Check which PDF files are missing (if any)"

    print("Checking for PDF files")
    print("----------------------")
    print("")

    # Iterate over papers
    num_found = 0
    num_missing = 0
    for paper in papers:

        # Skip some categories
        if paper["category"] in ["courses"]:
            continue

        # Generate name of PDF file
        pdffile = os.path.join("papers", paper["key"] + ".pdf")
        if os.path.isfile(pdffile):
            num_found += 1
            paper["pdf"] = pdffile
        else:
            print("Missing PDF file %s for paper: %s" % (pdffile, pstr(paper)))
            num_missing += 1
            paper["pdf"] = "missing"

    return (num_found, num_missing)



"This module implements input/output for pub."
from builtins import str

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-10-22 -- 2008-11-11"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Modified by Anders Logg 2012
# Modified by Benjamin Kehlet 2012
# Last changed: 2012-08-31

import re, sys

from publish.common import is_valid, ordered_attributes
from publish import config

# Pattern used for extracting the pub-categories
_category_list = "|".join([category for category in config.get("category_attributes")])
_category_pattern =  re.compile('^\*\s*(%s)\s*$' % _category_list, re.IGNORECASE)

# Pattern used for extracting title
_title_pattern = re.compile('^\*\*\s*(.*?)\s*$')

# Attributes to ignore
_ignores = ["invalid"]

def read(text):
    "Extract papers from text and return papers as a list of dictionaries."

    papers = []
    paper = None
    category = None
    title = None
    lines = text.split("\n")
    for line in lines:
        if sys.version[0] == '2':
            try:
                str(line)
            except UnicodeDecodeError as e:
                line = line.decode('utf-8')

        line = line.strip()

        # Look for start of category
        #match = _category_pattern.search(line)
        if line.startswith("* ") :

            # Extract category
            #groups = match.groups()
            #category = groups[0]
            category = line[2:]

            # Make sure every category is written in lower-case letters
            category = category.lower()

            continue

        # Look for start of paper
        match = _title_pattern.search(line)
        if not match is None:

            # Check that the category has been specified
            if category is None:
                raise RuntimeError("Found paper but category has not been specified.")

            # Extract title
            groups = match.groups()
            title = groups[0]

            # Save current paper if any
            if not paper is None:

                if "author" in paper:
                    paper["author"] = _extract_authors(paper, "author")
                if "editor" in paper:
                    paper["editor"] = _extract_authors(paper, "editor")

                if "allowed_duplicates" in paper :
                    # Split wrt whitespaces
                    paper["allowed_duplicates"] = paper['allowed_duplicates'].split()

                papers.append(paper)

            # Reset paper
            paper = {"title": title, "category": category}

            continue

        # Look for attribute
        if ":" in line:

            # Check that the title has been specified
            if title is None:
                raise RuntimeError("Found attribute but title has not been specified.")

            # Extract attribute and value
            attribute = line.split(":")[0].strip()
            value = ":".join(line.split(":")[1:]).strip()

            # Add attribute-value pair to paper
            if not attribute in _ignores:
                paper[attribute] = value

    # Save last paper if any
    if not paper is None:

        if "author" in paper:
            paper["author"] = _extract_authors(paper, "author")
        if "editor" in paper:
            paper["editor"] = _extract_authors(paper, "editor")

        papers.append(paper)

    return papers

def write(papers):
    "Format the given list of papers in the pub format."

    text = []

    if config.get("use_standard_categories") :
        categories = config.get("categories")
    else :
        categories = set()
        for paper in papers :
            categories.add(paper["category"])
        categories = list(categories)
        categories.sort()


    # Iterate over categories
    for category in categories:

        # Extract papers in category
        category_papers = [paper for paper in papers
                           if paper["category"] == category]
        if len(category_papers) == 0:
            continue

        # Write category
        text.append("* %s\n" % category)

        # Iterate over papers in category
        for paper in category_papers:

            # Write title
            if "title" in paper:
                title = paper["title"]
            else:
                title = "missing"
            try:
                text.append("** %s\n" % title)
            except UnicodeDecodeError as e:
                text.append("** %s\n" % title.decode('utf-8'))

            # Write attributes
            text.append(write_paper(paper, ["category", "title"] + _ignores))

    return "".join(text)

def write_paper(paper, ignores=[]):
    "Format given paper in the pub format"

    text = []

    # Extract which attributes to write (first required, then others)
    attributes = ordered_attributes(paper, ignores)

    # Make correct indentation for each attribute-value pair
    max_attr = max([len(attribute) for attribute in attributes])

    # Write attribute-value pairs
    for attribute in attributes:
        # Compute indendation
        indentation = " " * (max_attr - len(attribute))

        # Make sure we may write invalid papers with missing attributes
        if not attribute in paper:
            value = "missing"
        elif attribute == "author":
            value = ", ".join(paper["author"])
        elif attribute == "editor":
            value = ", ".join(paper["editor"])
        elif attribute == "allowed_duplicates" :
            value = " ".join(paper["allowed_duplicates"])
        else:
            value = paper[attribute]

        # Write attribute-value pair
        text.append("   %s: %s%s\n" % (attribute, indentation, value))

    return "".join(text)

def write_diff(paper0, paper1):
    "Write readable diff between papers"

    text = ""

    # Generate text for papers
    s0 = write_paper(paper0)
    s1 = write_paper(paper1)

    text += "First paper:\n\n"

    # Mark differences for first paper
    for line in s0.split("\n"):
        if not ":" in line: continue
        attribute = line.split(":")[0].strip()
        if attribute in paper1 and paper0[attribute] == paper1[attribute]:
            text += "    " + line.strip() + "\n"
        else:
            text += "--> " + line.strip() + "\n"

    text += "\nSecond paper:\n\n"

    # Mark differences for second paper
    for line in s1.split("\n"):
        if not ":" in line: continue
        attribute = line.split(":")[0].strip()
        if attribute in paper0 and paper0[attribute] == paper1[attribute]:
            text += "    " + line.strip() + "\n"
        else:
            text += "--> " + line.strip() + "\n"

    return text

def _extract_authors(paper, attribute):
    "Extract authors as tuple from string"

    author = paper[attribute]
    names = author.split(",")

    authors = []
    for name in names:
        name = name.strip()

        # Add missing . for initials and cleanup spaces
        words = name.split(" ")
        new_words = []
        for word in words:
            word = word.strip()
            if len(word) == 1:
                new_words.append(word + ".")
            elif len(word) > 1:
                new_words.append(word)
        name = " ".join(new_words)

        authors.append(name)

    return authors



# -*- coding: utf-8 -*-
"This module implements output for reStructuredText."


__author__ = "Anders Logg <logg@simula.no>"
__date__ = "2011-12-05 -- 2011-12-05"
__copyright__ = "Copyright (C) 2011 Anders Logg"
__license__  = "GNU GPL version 3 or any later version"

from publish import config

def write(papers, sort_func=None):
    "Format the given list of papers in HTML format."

    text = ""
    prefix = ""

    # Get the categories.
    # Assume unique names
    categories = config.get("categories")
    category_headings = config.get("category_headings")
    skip_categories = config.get("skip_categories")

    # Get formatting rule
    rst_format = config.get("rst_format")

    # Get PDF directory
    pdf_dir = config.get("pdf_dir")

    # Iterate over categories
    current_paper = 0
    if skip_categories:

        # Iterate over papers
        for paper in papers:

            # Format paper entry
            category = paper["category"]
            paper_entry = rst_format[category](paper)

            # Filter entry from special characters
            paper_entry = _filter(paper_entry)

            # Set directory to papers (a bit of a hack)
            paper_entry = paper_entry.replace("papers/", pdf_dir + "/")

            # Write entry for paper
            text += paper_entry + "\n"

    else:

        # Iterate over categories
        for category in categories:

            # Extract papers in category
            category_papers = [paper for paper in papers if paper["category"] == category]
            if len(category_papers) == 0:
                continue

            # Sort the list
            if sort_func is not None :
                category_papers.sort(sort_func)

            # Write category
            heading = category_headings[category]
            text += "%s\n" % heading
            text += "="*len(heading) + "\n\n"

            # Iterate over papers in category
            for paper in category_papers:

                # Format paper entry
                paper_entry = rst_format[category](paper)

                # Filter entry from special characters
                paper_entry = _filter(paper_entry)

                # Set directory to papers (a bit of a hack)
                paper_entry = paper_entry.replace("papers/", pdf_dir + "/")

                # Write entry for paper
                text += paper_entry + "\n"

    return text

def _filter(s):
    "Filter string for special characters."

    # List of replacements
    replacements = [("--", "&ndash;"),
                    ("$", ""),
                    ("\\mathrm", ""),
                    ('\\aa', 'å'),
                    ('\\AA', 'ä'),
                    ('\\"a', 'ä'),
                    ('\\"A', 'Ä'),
                    ('\\"o', 'ö'),
                    ('\\"O', 'Ö'),
                    ('\\o',  'ø'),
                    ('\\O',  'Ø'),
                    ('\\ae', 'æ'),
                    ('\\AE', 'Æ'),
                    ('\\&',   '&amp;'),
                    ('\\textonesuperior', "<sup>1</sup>"),
                    ('\\textendash', "&ndash;"),
                    ('\\textemdash', "&mdash;"),
                    ('\\textasciiacute', "&acute;"),
                    ('\\\'a', "&acute;"),
                    ('\\"u', "&Uuml;"),
                    ('\\epsilon', "&epsilon;"),
                    ('\\\'\i', "&iacute;"),
                    ('\\textquoteright', "&apos;") # Is this correct?
                    ]

    # Iterate over replacements
    for (a, b) in replacements:
        s = s.replace(a, b)

    return s



"This module implements input/output for BibTeX."
from __future__ import print_function
from builtins import str
from builtins import input
from builtins import range

__author__    = "Anna Logg (anna@loggsystems.se)"
__date__      = "2008-10-05 -- 2008-11-11"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__   = "GNU GPL version 3 or any later version"

# Modified by Benjamin Kehlet 2011
# Modified by Anders Logg 2012
# Last changed: 2012-01-30

import re

from publish.common import pstr, ordered_attributes
from publish import config
from publish.exceptions import ParseException

# Pattern used for extracting the BibTeX-fields
_entry_list = "|".join([entry_type for entry_type in config.get("entrytype_attributes")])
_entry_pattern = re.compile('^(%s)\s*{' % _entry_list, re.IGNORECASE)

# Pattern used for extracting everything before and after the "="-sign
#_block_pattern = re.compile('\s*(.*?)\s*=\s*{(.*?)}\s*,')

# Pattern used for extracting attribute
_attribute_pattern = re.compile('\s*,*(.*?)\s*=\s*')

# Attributes to ignore
_ignores = ["invalid", "pdf", "status", "category"]

def read(text):
    text = text.strip()
    "Extract papers from text and return papers as a list of dictionaries."

    print("")
    print("Importing papers from BibTeX")
    print("----------------------------")
    print("")

    papers = []
    position = 0

    while position < len(text) :
        current_paper = {}

        position = _skip_spaces(position, text)

        # Look for '@' at start of paper
        if not text[position] == "@" :
            raise ParseException("BibTeX parse error expected '@' near '%s'" % _get_line(position, text))

        position += 1

        # Extract entry-type
        entrytype_orig = text[position:position+text[position:].find("{")].strip()
        entrytype = entrytype_orig.lower()  # bring everything to lower case
        #match = re.search(_entry_pattern, text[position:])  # old

        if config.get("use_standard_categories") and entrytype not in config.get("entrytype_attributes") :
            #if not match :
            msg  = "Parse error in BibTeX file\n%s\n" % _get_line(position, text)
            msg += 'Found entry "%s"; allowed entry types are: %s' % (entrytype_orig, ", ".join(config.get("entrytype_attributes")))
            raise ParseException(msg)

        # Make sure every entry-type is written in lower-case letters
        #current_paper["entrytype"] = match.group(1).lower()
        current_paper["entrytype"] = entrytype

        position += len(current_paper["entrytype"])
        position = _skip_spaces(position, text)

        if not text[position] == "{" :
            raise ParseException("Bibtex parse error. Expected '{' after entrytype. Got '%s'\n%s" % (text[position], _get_line(position, text)))

        position += 1

        # Find the ',' which separates the key from the attributes
        index = text[position:].find(",")

        current_paper["key"] = text[position:position+index].strip()

        # skip through the key and the following ','
        position += index+1
        position = _skip_spaces(position, text)

        # collect all the attributes
        while True :
            position, attr_key, attr_value = _parse_attribute(position, text)
            position = _skip_spaces(position, text)

            # Bibtex attribute "key" is legal and needed for sorting if
            # author is missing (typical in software entries). Let
            # Bibtex "key" correspond to "sortkey" in publish
            if attr_key == "key":
                attr_key = "sortkey"

            if attr_key in current_paper :
                raise ParseException("Paper with key '%s' has double declared attribute: '%s'" % (current_paper["key"], attr_key))

            current_paper[attr_key] = attr_value

            if text[position] != "," and text[position] != "}" :
                raise ParseException("Unexpected character '%s' at %s" % (text[position], _get_line(position, text)))

            position = _skip_spaces(position ,text)
            if text[position] == "," :
                position += 1
                position = _skip_spaces(position, text)

            # Note that a comma is allowed after the last attribute
            if text[position] == "}" :
                break


        position += 1
        position = _skip_spaces(position, text)

        # Done with parsing the paper. Now validate the collected data

        # Check that paper has all required attributes
        if config.get("use_standard_categories") :
            _check_paper(current_paper)
        else :
            if "status" not in current_paper :
                current_paper["status"] = "published"

        # Extract category
        if config.get("use_standard_categories") :
            current_paper["category"] = _extract_category(current_paper)
        else :
            current_paper["category"] = current_paper["entrytype"]

        # Extract authors as tuple from string
        if "author" in current_paper:
            _extract_authors(current_paper, "author")

        # Extract editors as tuple from string
        if "editor" in current_paper:
            _extract_authors(current_paper, "editor")

        papers.append(current_paper)

    # Return list of papers
    return papers

def _parse_attribute(position, text) :
    eq_pos = text[position:].find("=")

    if eq_pos < 1 :
      raise ParseException("Bibtex parse error, expected attribute=value near '%s'" % _get_line(position, text))

    key = text[position:position+eq_pos].strip()

    position += eq_pos+1

    position, value = _parse_attribute_value(position, text)
    return (position, key, value)

def _parse_attribute_value(position, text) :
    position = _skip_spaces(position, text)

    if not text[position] == "{" :
        raise ParseException("Expected '{' near '%s'" % _get_line(position, text))

    position += 1

    # Look for the end point, by counting the braces
    num_left_braces = 1
    num_right_braces= 0
    end_pos = position
    for end in range(position, len(text)):

        # Count braces
        if text[end] == "{" and not text[end - 1] == "\\" :
            num_left_braces += 1
        elif text[end] == "}" and not text[end -1] == "\\" :
            num_right_braces += 1

        # Found end of paper, done
        if num_left_braces == num_right_braces:
            break

    #orig: return (end+1, text[position:end].strip())
    value = text[position:end].strip().replace('\n', ' ')
    value = re.sub(r' +', ' ', value)
    return (end+1, value)


def write(papers):
    "Format the given list of papers in the BibTeX format."

    text = ""

    for (i, paper) in enumerate(papers):
        entry_type = config.get("category2entrytype")[paper["category"]]
        if "key" in paper:
            key = paper["key"]
        else:
            key = "paper%d" % i
        text += "@%s{%s,\n" % (entry_type, key)
        for attribute in ordered_attributes(paper, _ignores):
            if attribute in ("entrytype", "key"):
                continue
            if attribute == "sortkey":
                attribute = "key"  # sortkey becomes key in Bibtex
            if attribute == "author":
                value = " and ".join(paper["author"])
            elif attribute == "editor":
                value = " and ".join(paper["editor"])
            else:
                value = str(paper[attribute])
            text += "  %s = {%s},\n" % (attribute, value)
        text += "}\n"
        if not paper == papers[-1]:
            text += "\n"

    return text

def _parse_paper(block, value_start, value_end):
    "Parse paper attributes"

    # Check lists of positions, should be of equal size
    if not len(value_start) == len(value_end):
        raise RuntimeError("Syntax error in BibTeX file, unbalanced braces.")

    # Extract attributes and values
    paper = {}
    for i in range(len(value_start)):

        # Extract attribute
        if i == 0:
            attribute = block[0:value_start[0]]
        else:
            attribute = block[value_end[i - 1] + 1:value_start[i]]
        match = _attribute_pattern.search(attribute)
        if match is None:
            raise RuntimeError("Syntax error in BibTeX file, missing attribute name.")
        groups = match.groups()
        attribute = groups[0].lower()

        # Extract value
        value = block[value_start[i] + 1:value_end[i]].replace("\r", " ").replace("\n", " ").strip()

        # Set attribute and value
        paper[attribute] = value

    # Return a dictionary containing information about one paper
    return paper

def _check_paper(paper):
    "Check required attributes"

    # TODO: Do this during parsing, so we can give error messages with line number and text

    print("Found paper: %s" % pstr(paper))

    invalid = False

    # Check that paper has all required attributes
    entry_type = paper["entrytype"]
    key = paper["key"]
    attributes = config.get("entrytype_attributes")[entry_type]
    for attribute in attributes:
        # Check if the required field is a tuple and at least one field is used
        if isinstance(attribute, tuple):
            if not len([f for f in attribute if f in paper]) >= 1:
                print('  Missing required attribute(s) "%s" for paper "%s"' % ('"/"'.join(attribute), key))
                invalid = True
        elif not attribute in paper:
            print('  Missing required attribute "%s" for paper "%s"' % (attribute, key))
            invalid = True

    if invalid:
        paper["invalid"] = True
        print("  Skipping paper. Correct the above error(s) and import the paper again.")
        if not config.get("autofix"):
            input("  Press return to continue.")

def _extract_authors(paper, attribute):
    "Extract authors as tuple from string"

    names = paper[attribute].split(" and ")

    authors = []
    for name in names:
        name = name.strip()

        # Handle case Last, First
        if "," in name:
            words = name.split(",")
            if not len(words) == 2:
                paper["invalid"] = True
                print("  Incorrectly formatted author string:", name)
                if config.get("autofix"):
                    print("  Skipping paper.")
                else:
                    input("  Skipping paper, press return to continue.")
            words.reverse()
            name = " ".join([w.strip() for w in words])

        # Add missing . for initials and cleanup spaces
        words = name.split(" ")
        new_words = []
        for word in words:
            word = word.strip()
            if len(word) == 1:
                new_words.append(word + ".")
            elif len(word) > 1:
                new_words.append(word)
        name = " ".join(new_words)

        authors.append(name)

    paper[attribute] = authors

def _extract_category(paper):
    "Extract category for paper"

    # Check if category is supported
    entry_type = paper["entrytype"]
    entrytype2category = config.get("entrytype2category")
    if not entry_type in entrytype2category:
        raise RuntimeError('Entry type "%s" not supported.' % entry_type)

    # Check default mapping
    category = entrytype2category[entry_type]
    if not category is None:
        return category

    # Special case: book or edited
    if entry_type == "book":
        if "editor" in paper:
            paper["author"] = paper["editor"]
            del paper["editor"]
            return "edited"
        else:
            return "books"

    # Special case: phdthesis
    if entry_type == "phdthesis":
        paper["thesistype"] = "phd"
        return "theses"

    # Special case: mscthesis
    if entry_type == "mastersthesis":
        paper["thesistype"] = "msc"
        return "theses"

    # Special case: misc
    if entry_type == "misc":
        if "code" in paper:
            return "courses"
        elif "meeting" in paper:
            return "talks"
        else:
            return "misc"

    raise RuntimeError("Unhandled special case for BibTeX entry type.")


def _skip_spaces(position, text) :
    "Get position of first character after position, not beeing a white space"
    while  position < len(text)-1 and text[position].isspace() :
        position += 1

    return position

def _get_line(position, text) :
    "Get the string with linenumber and text of the line that text[position] belongs to"

    # count newlines before position to get the linenumber
    newlines = text.count("\n", 0, position) + 1

    start = text[:position].rfind("\n")
    end   = text[position:].find("\n")
    return "(%d) : %s" % (newlines, text[start:position+end].strip())



"This module implements some useful algorithms."

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-02 -- 2008-11-02"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

import Levenshtein as levenshtein

def distance(s0, s1):
    "Compute the relative edit distance between the two strings."

    d  = float(levenshtein.distance(s0, s1))
    d /= max(len(s0), len(s1), 1)

    return d



"This module implements input/output for LaTeX."

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-06 -- 2008-11-12"
__copyright__ = "Copyright (C) 2008 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

from publish import config


def write(papers, sort_func=None):
    "Format the given list of papers in the LaTeX format."

    text = ""

    # Get formatting rule
    latex_format = config.get("latex_format")
    compact = config.get("compact")
    global_numbering = config.get("global_numbering")
    use_labels = config.get("use_labels")
    category_labels = config.get("category_labels")

    # Start of LaTeX paper
    if global_numbering and not compact:
        text+= "\\begin{thebibliography}{99}\n"

    # Iterate over categories
    categories = config.get("categories")
    current_paper = 0
    for category in categories:
        # Extract papers in category
        category_papers = [paper for paper in papers if paper["category"] == category]
        if len(category_papers) == 0:
            continue

        # Sort the list
        if sort_func is not None :
          category_papers.sort(key=sort_func)

        # Write category
        category_headings = config.get("category_headings")
        if compact:
            text += "\n\\textit{%s}\n\n" % category_headings[category]
        else:
            text += "\\subsection*{%s}\n" % category_headings[category]

        if not global_numbering and not compact :
            text += "\\begin{thebibliography}{99}\n"

        # Iterate over papers in category
        for (counter, paper) in enumerate(category_papers):

            # Get key (or generate key)
            if "key" in paper:
                key = paper["key"]
            else:
                key = "paper%d" % current_paper

            entry_text = latex_format[category](paper)

            # Set bibitem key
            if use_labels:
                label = "[%s%d]" % (category_labels[category], counter + 1)
            else:
                label = ""

            # Write each paper as bibitem
            if compact:
                text += "[%d] %s\\\\[1ex]\n" % (current_paper, entry_text)
            else:
                text += "\\bibitem%s{%s} {%s}\n" % (label, key, entry_text)

            current_paper += 1

        if not global_numbering and not compact :
            text += "\\end{thebibliography}\n\n"


    if global_numbering and not compact:
        text += "\\end{thebibliography}"

    return text



"This module implements data export to different file formats."
from __future__ import print_function
from __future__ import absolute_import

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-11 -- 2009-08-30"
__copyright__ = "Copyright (C) 2008-2009 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Last modified: 2011-12-05

# Modified by Anders Logg, 2009-2011.

from publish.importing import read_database
from publish import config
from .formats import bibtex, latex, pub, pdf, html, rst, graphml
from .validation import validate_papers
from .filtering import filter_papers
from .log import print_summary

def export_file(filename, filters=[]):
    "Export data into desired file format"

    # Make sure we don't overwrite the database
    database_filename = config.get("database_filename")
    if filename == database_filename:
        raise RuntimeError('Papers cannot be exported to the default database ("%s").' % database_filename)

    # Read database
    database_papers = read_database(database_filename)

    # Why should the database be validated on export?
    #(valid_papers, invalid_papers) = validate_papers(database_papers)

    # Filter papers
    filtered_papers = filter_papers(database_papers, filters)

    # Get the filename suffix
    suffix = filename.split(".")[-1]

    # Choose format based on suffix
    if suffix in ("bib", "bibtex"):
        write = bibtex.write
    elif suffix == "pub":
        write = pub.write
    elif suffix == "tex":
        write = latex.write
    elif suffix == "pdf":
        write = pdf.write
    elif suffix == "html":
        write = html.write
    elif suffix == "rst":
        write = rst.write
    elif suffix == "graphml" :
        write = graphml.write
    else:
        raise RuntimeError("Unknown file format.")

    # Open and read file
    text = write(filtered_papers)
    file = open(filename, "w")
    try:
        file.write(text)
    except UnicodeEncodeError as e:
        file.write(text.encode('utf-8'))
    except Exception as e:
        raise RuntimeError('Unable to write file "%s" (exception %s: %s)'
                           % (filename, type(e), e))
    file.close()

    # Print summary
    print_summary(filtered_papers)
    print("")
    print("Exported %d paper(s) to %s." % (len(filtered_papers), filename))



"This module handles configuration parameters."
from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str

__author__ = "Anna Logg (anna@loggsystems.se)"
__date__ = "2008-11-09 -- 2009-07-31"
__copyright__ = "Copyright (C) 2008-2009 Anna Logg"
__license__  = "GNU GPL version 3 or any later version"

# Modified by Anders Logg 2009-2012
# Modified by Benjamin Kehlet 2012
# Last modified: 2012-06-02

from os.path import isfile
import builtins, sys
data = None

def get(key):
    "Return data/parameter value for given key"

    # Initialize parameters if necessary
    if data is None:
        init()

    # Check key
    if not key in data:
        raise RuntimeError("Unknown parameter key: %s"% str(key))

    return data[key]

def set(key, value):
    "Set data/parameter value for given key"

    # Check key
    if not key in data:
        raise RuntimeError("Unknown parameter key: %s"% str(key))

    # Set value
    data[key] = value

def set_from_string(key, value):
    "Set data/parameter value from string for given key"
    if value == "yes":
        set(key, True)
    elif value == "no":
        set(key, False)
    else:
        #set(key, eval(value))
        set(key, value)

def has_key(key):
    "Check if parameter exist for key"
    if data is None:
        init()

    return key in data

def init():

    global data
    data = {}

    try:
        # Import from user's configuration file, which may
        # perform from publish.config.default import * first
        # and then overrides and/or adds information, or
        # just redefine variables and extending lists and dicts.
        import publish_config as user_config
        path = str(sys.modules['publish_config']).split('from')[1].split("'")[1]
        if '/tests/local_config_publish_import' not in path:
            print('user configuration data for publish is imported from\n    ', path)
    except ImportError:
        user_config = None

    from . import defaults

    if user_config is not None:
        if hasattr(user_config, 'MARKER_FOR_IMPORT_0123456789'):
            # User's publish_config.py has done a
            #   from publish.defaults.config import *
            # and represents the uninion of defaults and user's data
            defaults = user_config
        else:
            # Add user's data to those in defaults
            for name in dir(defaults):
                var = getattr(defaults, name)
                if name.startswith('_'):
                    # Drop "private" data
                    continue
                if not hasattr(user_config, name):
                    # User has not set this variable
                    continue
                if isinstance(var, (bool,int,float,str)):
                    # Override basic variable with user's value
                    setattr(defaults, name,
                            getattr(user_config, name))
                elif isinstance(var, (tuple,list)):
                    # Extend defaults' var with user's
                    setattr(defaults, name,
                            getattr(user_config, name) + var)
                    # (we do obj = obj + obj since it works with
                    # pure tuples and list, alternative is in-place
                    # var.extend(getattr(user_config, name))
                    # but that works only for lists)
                elif isinstance(var, dict):
                    # inplace change (no assignment)
                    var.update(getattr(user_config, name))
                else:
                    raise TypeError(
                        'defaults.py variable "%s" is of a type not '
                        'handled by the code in config/__init__.py')
                conf = defaults

    # Backward compatibility of names
    general = defaults
    attributes = defaults
    capitalization = defaults
    schools = defaults
    publishers = defaults
    typos = defaults
    institutions = defaults
    journals = defaults

    # Import parameters from general
    data["database_filename"]        = general.database_filename
    data["local_venues_filename"]    = general.local_venues_filename
    data["authornames_filename"]     = general.authornames_filename
    data["invalid_filename_prefix"]  = general.invalid_filename_prefix
    data["matching_distance_strong"] = general.matching_distance_strong
    data["matching_distance_weak"]   = general.matching_distance_weak
    data["autofix"]                  = general.autofix
    data["debug"]                    = general.debug
    data["pdf_viewer"]               = general.pdf_viewer
    data["view_pdf"]                 = general.view_pdf
    data["pdf_dir"]                  = general.pdf_dir
    data["headline"]                 = general.headline
    data["repeat_headline"]          = general.repeat_headline
    data["compact"]                  = general.compact
    data["require_page_range"]       = general.require_page_range
    data["page_separator"]           = general.page_separator
    data["html_class_prefix"]        = general.html_class_prefix
    data["html_add_internal_links"]  = general.html_add_internal_links
    data["global_numbering"]         = general.global_numbering
    data["use_labels"]               = general.use_labels
    data["skip_categories"]          = general.skip_categories
    data["talks_dont_duplicate"]     = general.talks_dont_duplicate
    data["use_standard_categories"]  = general.use_standard_categories
    data["simple_import"]            = general.simple_import

    # Import parameters from capitalization
    data["lowercase"] = capitalization.lowercase
    data["uppercase"] = capitalization.uppercase

    # Import parameters from typos
    data["typos"] = typos.typos

    # Import parameters from attributes
    data["categories"]           = attributes.categories
    data["category_headings"]    = attributes.category_headings
    data["category_labels"]      = attributes.category_labels
    data["category_attributes"]  = attributes.category_attributes
    data["category_venues"]      = attributes.category_venues
    data["ordered_attributes"]   = attributes.ordered_attributes
    data["entrytypes"]           = attributes.entrytypes
    data["entrytype_attributes"] = attributes.entrytype_attributes
    data["entrytype2category"]   = attributes.entrytype2category
    data["category2entrytype"]   = attributes.category2entrytype
    data["thesistype_strings"]   = attributes.thesistype_strings

    # Import parameters from formatting
    from . import formatting
    data["latex_format"] = formatting.latex_format
    data["html_format"] = formatting.html_format
    data["rst_format"] = formatting.rst_format
    data["mark_author"] = builtins.set()
    data["use_textsc"]  = True

    # Import parameters from institutions
    data["institutions"] = list(institutions.institutions)

    # Import parameters from schools
    data["schools"] = list(schools.schools)

    # Import parameters from publishers
    data["publishers"] = list(publishers.publishers)

    # Empty list of meetings (could be extended in meetings.py)
    data["meetings"] = []

    # Import parameters from journals
    # Create list of all journals (including both short and long names)
    journal_list = [short for (short, int, issn) in journals.journals] + \
                   [int  for (short, int, issn) in journals.journals]
    data["journals"] = journal_list

    # Create mapping from long to short journal names
    long2short = {}
    for (short, int, issn) in journals.journals:
        long2short[int] = short
    data["long2short"] = long2short

    # Create mapping from short to long journal names
    short2long = {}
    for (short, int, issn) in journals.journals:
        short2long[short] = int
    data["short2long"] = short2long

    # Create mapping from long name to issn number
    long2issn = {}
    for (short, int, issn) in journals.journals:
        long2issn[int] = issn
    data["long2issn"] = long2issn

    # Read local venues
    _read_uservenues(data["local_venues_filename"],
                     data["journals"],
                     data["publishers"],
                     data["schools"],
                     data["institutions"],
                     data["meetings"])

    # Read list of author names
    data["allowed_author_names"] = _read_author_names(data["authornames_filename"])

def _read_uservenues(filename, journals, publishers, schools, institutions, meetings):
    "Read venues from file"

    # Check for file
    if not isfile(filename):
        return

    # Open and read file
    try:
        file = open(filename, "r")
        text = file.read()
        file.close()
    except:
        raise RuntimeError('Unable to read local venues from file "%s".' % filename)

    # Parse file
    lines = text.split("\n")
    for line in lines:

        if not ":" in line:
            continue

        venue_type = line.split(":")[0].strip()
        venue_name = ":".join(line.split(":")[1:]).strip()

        if venue_type == "journal":
            journals.append(venue_name)
        elif venue_type == "publisher":
            publishers.append(venue_name)
        elif venue_type == "school":
            schools.append(venue_name)
        elif venue_type == "institution":
            institutions.append(venue_name)
        elif venue_type == "meeting":
            meetings.append(venue_name)
        else:
            print('Unknown venue type: "%s"' % venue_type)

def _read_author_names(filename) :
    # Note: Name clash with this modules set function
    authors = builtins.set()
    if isfile(filename) :
        with open(filename, "r") as f :
            for line in f :
                author = line.strip()
                if not author :
                    continue
                authors.add(author)
        return authors
    else :
        return None
